# Copyright 2026 EcoFuture Technology Services LLC and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from django.db import models

from bazis.core.models_abstract import DtMixin, JsonApiMixin, UuidMixin
from bazis.core.utils.orm import (
    FieldJson,
    FieldSubAggr,
    calc_property,
)


class Gender(models.TextChoices):
    MALE = (
        'm',
        'Male',
    )
    FEMALE = (
        'f',
        'Female',
    )


class People(DtMixin, UuidMixin, JsonApiMixin):
    name = models.CharField(max_length=255)
    email = models.CharField(
        'Email',
        max_length=255,
    )
    gender = models.CharField(
        'State',
        max_length=1,
        choices=Gender,
    )


class Category(DtMixin, UuidMixin, JsonApiMixin):
    name = models.CharField(max_length=255)
    comment = models.CharField(max_length=255)


class Article(DtMixin, UuidMixin, JsonApiMixin):
    author = models.ManyToManyField(
        People,
        related_name='articles',
        blank=True,
    )
    title = models.CharField(
        'Title',
        max_length=25,
    )
    body = models.CharField(
        'Title',
        max_length=250,
    )
    category = models.ForeignKey(
        'Category', on_delete=models.CASCADE, related_name='articles', null=True
    )

    @calc_property(
        [
            FieldJson(
                source='author',
                fields=[
                    'id',
                    'name',
                ],
            ),
        ]
    )
    def author_detail(self) -> list:
        return [
            {
                'id': author['id'],
                'child_name': author['name'],
            }
            for author in self._author
        ]

    @calc_property(
        [
            FieldSubAggr(
                'author',
                func='Count',
                alias='author_count',
            ),
        ]
    )
    def author_count(self) -> int:
        return self.author_count

    @property
    def some_count_property(self) -> int:
        return 1000



# Copyright 2026 EcoFuture Technology Services LLC and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os

from django.conf import settings
from django.core.management.base import BaseCommand
from django.utils.translation import get_language

import jsonref


def get_definitions(schema):
    if settings.BAZIS_SCHEMA_WITHOUT_REF:
        schema = jsonref.replace_refs(schema)

    definitions = schema['components']['schemas']

    if settings.BAZIS_SCHEMA_WITHOUT_REF:
        return {key: it for key, it in definitions.items() if not key.startswith('_')}
    else:
        return definitions


class Command(BaseCommand):
    """
    The command creates OpenAPI schemas for the current language
    and saves them to a JSON file in the static directory.

    Tags: RAG
    """

    def handle(self, **kwargs):
        from bazis.core.app import app
        from bazis.core.router import router
        from bazis.core.routing import BazisRoute

        os.makedirs(settings.STATIC_ROOT, exist_ok=True)

        router.routes_cast(BazisRoute)
        app.include_router(router)

        with open(os.path.join(settings.STATIC_ROOT, f'schemas_{get_language()}.json'), 'w') as fp:
            json.dump(get_definitions(app.openapi()), fp, ensure_ascii=False)



# Copyright 2026 EcoFuture Technology Services LLC and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Context

Using context and query, you can configure automatic filtering of data obtained from related tables.
"""

import pytest
from bazis_test_utils.utils import get_api_client

from tests.utils.assert_sql import assert_sql_query, get_sql_query


@pytest.mark.django_db(transaction=True)
def test_context(sample_app, dynamic_vehicle_data):
    """
    Hierarchy of related tables.

    A calculated field for a vehicle (Vehicle model) with "_organization" specified in context and
    org_owner=F('_organization') in query for data from the related table with trips (CarrierTasks model).

    .. code-block:: python

        @calc_property([
            FieldDynamic(source='carrier_tasks',
                         fields=['id', 'dt_start', 'dt_finish', 'fact_waste_weight', 'vehicle_id'],
                         context=['_organization'],
                         query=Q(org_owner=F('_organization')))
            .add_nested([
                FieldDynamic(
                    source='driver',
                    fields=['contact_phone'])
                .add_nested([
                    FieldDynamic(source="org_owner",
                                 fields=["name"])
                ])
            ])
        ])
        def carrier_tasks_context(self, dc: DependsCalc) -> list:
            return [
                {
                    'id': task.id,
                    'dt_start': task.dt_start,
                    'dt_finish': task.dt_finish,
                    'fact_waste_weight': task.fact_waste_weight,
                    'driver_phone': task.driver[0].contact_phone,
                    'organization': task.driver[0].org_owner[0].name,
                }
                for task in dc.data.carrier_tasks
            ]
    """

    organization = dynamic_vehicle_data['organization']

    url = '/api/v1/context/dynamic/vehicle/'

    response = get_api_client(sample_app).get(url)

    assert response.status_code == 200

    sql_query = get_sql_query()
    expected_sql_template = """
        SELECT "dynamic_vehicle"."dt_created",
               "dynamic_vehicle"."dt_updated",
               "dynamic_vehicle"."id",
               "dynamic_vehicle"."vehicle_model_id",
               "dynamic_vehicle"."gnum",
               ARRAY
          (SELECT JSONB_BUILD_OBJECT(('id')::text, W0."id", ('dt_start')::text, W0."dt_start", ('dt_finish')::text, W0."dt_finish", ('fact_waste_weight')::text, W0."fact_waste_weight", ('_driver')::text, ARRAY
                                       (SELECT JSONB_BUILD_OBJECT(('contact_phone')::text, V0."contact_phone", ('_org_owner')::text, ARRAY
                                                                    (SELECT JSONB_BUILD_OBJECT(('name')::text, U0."name") AS "json"
                                                                     FROM "dynamic_organization" U0
                                                                     WHERE U0."id" = (V0."org_owner_id"))) AS "json"
                                        FROM "dynamic_driver" V0
                                        WHERE V0."id" = (W0."driver_id"))) AS "json"
           FROM "dynamic_carriertask" W0
           WHERE (W0."vehicle_id" = ("dynamic_vehicle"."id")
                  AND W0."org_owner_id" = '3fa85f64-5717-4562-b3fc-2c963f66afa6')) AS "_carrier_tasks"
        FROM "dynamic_vehicle"
        LIMIT 20;"""

    assert_sql_query(expected_sql_template, sql_query, organization.id)



# Copyright 2026 EcoFuture Technology Services LLC and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from django.conf import settings

import pytest
from bazis_test_utils.utils import get_api_client
from entity.models import (
    DependentEntity,
    ExtendedEntity,
    ParentEntity,
)

from tests import factories


@pytest.mark.django_db(transaction=True)
def test_routes(sample_app):
    # parent_entity
    # LIST
    factories.ParentEntityFactory.create_batch(50)

    response = get_api_client(sample_app).get('/api/v1/entity/parent_entity/')

    assert response.status_code == 200

    data = response.json()

    assert len(data['data']) == settings.BAZIS_API_PAGINATION_PAGE_SIZE_DEFAULT

    for it in data['data']:
        assert it['type'] == 'entity.parent_entity'
        assert it['bs:action'] == 'view'
        assert 'id' in it
        assert 'attributes' in it
        assert 'relationships' in it

        _attributes = it['attributes']
        _relationships = it['relationships']

        assert 'dt_created' in _attributes
        assert 'dt_updated' in _attributes
        assert 'name' in _attributes
        assert 'description' in _attributes
        assert 'is_active' in _attributes
        assert 'price' in _attributes
        assert 'dt_approved' in _attributes
        assert 'state' in _attributes
        assert 'field' in _attributes
        assert 'extended_entity_price' in _attributes
        assert 'active_children' in _attributes
        assert 'count_active_children' in _attributes
        assert 'has_inactive_children' in _attributes

        assert 'child_entities' in _relationships
        assert 'extended_entity' in _relationships
        assert 'dependent_entities' in _relationships

        _extended_entity = _relationships['extended_entity']
        _dependent_entities = _relationships['dependent_entities']
        _child_entities = _relationships['child_entities']

        assert _extended_entity['data']['id'] is not None
        assert _extended_entity['data']['type'] == 'entity.extended_entity'
        assert isinstance(_dependent_entities['data'], list)
        assert _dependent_entities['data'][0]['id'] is not None
        assert _dependent_entities['data'][0]['type'] == 'entity.dependent_entity'
        assert _child_entities['data'] == []

    # RETRIEVE
    parent_entity = ParentEntity.objects.first()
    child_entity = factories.ChildEntityFactory.create()
    parent_entity.child_entities.add(child_entity)
    parent_entity.refresh_from_db()

    response = get_api_client(sample_app).get(
        f'/api/v1/entity/parent_entity/{str(parent_entity.id)}/'
    )

    assert response.status_code == 200

    data = response.json()

    assert data['data']['type'] == 'entity.parent_entity'
    assert data['data']['bs:action'] == 'view'
    assert data['data']['id'] == str(parent_entity.id)

    assert data['data']['attributes']['dt_created'] == parent_entity.dt_created.isoformat().replace(
        '+00:00', 'Z'
    )
    assert data['data']['attributes']['dt_updated'] == parent_entity.dt_updated.isoformat().replace(
        '+00:00', 'Z'
    )
    assert data['data']['attributes']['name'] == parent_entity.name
    assert data['data']['attributes']['description'] == parent_entity.description
    assert data['data']['attributes']['is_active'] == parent_entity.is_active
    assert data['data']['attributes']['price'] == str(parent_entity.price)
    assert data['data']['attributes'][
        'dt_approved'
    ] == parent_entity.dt_approved.isoformat().replace('+00:00', 'Z')
    assert data['data']['attributes']['state'] == parent_entity.state
    assert data['data']['attributes']['field'] == parent_entity.field
    assert 'extended_entity_price' in data['data']['attributes']
    assert 'active_children' in data['data']['attributes']
    assert 'count_active_children' in data['data']['attributes']
    assert 'has_inactive_children' in data['data']['attributes']

    _relationships = data['data']['relationships']

    assert 'child_entities' in _relationships
    assert 'extended_entity' in _relationships
    assert 'dependent_entities' in _relationships

    _extended_entity = _relationships['extended_entity']
    _dependent_entities = _relationships['dependent_entities']
    _child_entities = _relationships['child_entities']

    assert _extended_entity['data']['id'] is not None
    assert _extended_entity['data']['type'] == 'entity.extended_entity'
    assert isinstance(_dependent_entities['data'], list)
    assert _dependent_entities['data'][0]['id'] is not None
    assert _dependent_entities['data'][0]['type'] == 'entity.dependent_entity'
    assert _child_entities['data'] is not None
    assert _child_entities['data'][0]['id'] == str(child_entity.id)
    assert _child_entities['data'][0]['type'] == 'entity.child_entity'

    # SCHEMA
    response = get_api_client(sample_app).get('/api/v1/entity/parent_entity/schema_list/')
    assert response.status_code == 200

    data = response.json()
    assert data['$defs'] is not None
    assert data['properties'] is not None
    assert data['properties']['data'] is not None
    assert data['properties']['meta'] is not None
    assert data['required'] is not None

    check_data = 0
    for schema in data['$defs'].keys():
        if (
            schema.endswith('Attributes__Attributes')
            or schema.endswith('Relationships__Data__child_entities')
            or schema.endswith('Relationships__Data__dependent_entities')
            or schema.endswith('Relationships__Data__extended_entity')
        ):
            check_data += 1
        if schema.endswith('Attributes__Attributes'):
            assert data['$defs'][schema]['properties']['name']['filterLabel'] == 'name'
            assert data['$defs'][schema]['properties']['name']['orderLabel'] == 'name'

    assert check_data == 4

    response = get_api_client(sample_app).get('/api/v1/entity/parent_entity/schema_create/')
    assert response.status_code == 200

    data = response.json()
    assert data['$defs'] is not None
    assert data['properties'] is not None
    assert data['properties']['data'] is not None
    assert data['properties']['meta'] is not None
    assert data['required'] is not None

    check_data = 0
    for schema in data['$defs'].keys():
        if (
            schema.endswith('Attributes__Attributes')
            or schema.endswith('Relationships__Data__child_entities')
            or schema.endswith('Relationships__Data__dependent_entities')
            or schema.endswith('Relationships__Data__extended_entity')
        ):
            check_data += 1
        if schema.endswith('Attributes__Attributes'):
            assert data['$defs'][schema]['properties']['price']['filterLabel'] == 'price'
            assert data['$defs'][schema]['properties']['price']['orderLabel'] == 'price'

    assert check_data == 4

    response = get_api_client(sample_app).get(
        f'/api/v1/entity/parent_entity/{str(parent_entity.id)}/schema_retrieve/'
    )
    assert response.status_code == 200

    data = response.json()
    assert data['$defs'] is not None
    assert data['properties'] is not None
    assert data['properties']['data'] is not None
    assert data['properties']['meta'] is not None
    assert data['required'] is not None

    check_data = 0
    for schema in data['$defs'].keys():
        if (
            schema.endswith('Attributes__Attributes')
            or schema.endswith('Relationships__Data__child_entities')
            or schema.endswith('Relationships__Data__dependent_entities')
            or schema.endswith('Relationships__Data__extended_entity')
        ):
            check_data += 1

        if schema.endswith('Attributes__Attributes'):
            assert data['$defs'][schema]['properties']['is_active']['filterLabel'] == 'is_active'
            assert data['$defs'][schema]['properties']['is_active']['orderLabel'] == 'is_active'

    assert check_data == 4

    response = get_api_client(sample_app).get(
        f'/api/v1/entity/parent_entity/{str(parent_entity.id)}/schema_update/'
    )
    assert response.status_code == 200

    data = response.json()
    assert data['$defs'] is not None
    assert data['properties'] is not None
    assert data['properties']['data'] is not None
    assert data['properties']['meta'] is not None
    assert data['required'] is not None

    check_data = 0
    for schema in data['$defs'].keys():
        if (
            schema.endswith('Attributes__Attributes')
            or schema.endswith('Relationships__Data__child_entities')
            or schema.endswith('Relationships__Data__dependent_entities')
            or schema.endswith('Relationships__Data__extended_entity')
        ):
            check_data += 1
        if schema.endswith('Attributes__Attributes'):
            assert data['$defs'][schema]['properties']['dt_updated']['filterLabel'] == 'dt_updated'
            assert data['$defs'][schema]['properties']['dt_updated']['orderLabel'] == 'dt_updated'

    assert check_data == 4

    # child_entity
    # LIST
    factories.ChildEntityFactory.create_batch(20)

    response = get_api_client(sample_app).get('/api/v1/entity/child_entity/')

    assert response.status_code == 200

    data = response.json()

    assert len(data['data']) == settings.BAZIS_API_PAGINATION_PAGE_SIZE_DEFAULT

    for it in data['data']:
        assert it['type'] == 'entity.child_entity'
        assert it['bs:action'] == 'view'
        assert 'id' in it
        assert 'attributes' in it
        assert 'relationships' in it

        _attributes = it['attributes']
        _relationships = it['relationships']

        assert 'dt_created' in _attributes
        assert 'dt_updated' in _attributes
        assert 'child_name' in _attributes
        assert 'child_description' in _attributes
        assert 'child_is_active' in _attributes
        assert 'child_price' in _attributes
        assert 'child_dt_approved' in _attributes

        assert 'parent_entities' in _relationships

    # RETRIEVE
    child_entity.refresh_from_db()
    response = get_api_client(sample_app).get(
        f'/api/v1/entity/child_entity/{str(child_entity.id)}/'
    )

    assert response.status_code == 200

    data = response.json()

    assert data['data']['type'] == 'entity.child_entity'
    assert data['data']['bs:action'] == 'view'
    assert data['data']['id'] == str(child_entity.id)

    assert data['data']['attributes']['dt_created'] == child_entity.dt_created.isoformat().replace(
        '+00:00', 'Z'
    )
    assert data['data']['attributes']['dt_updated'] == child_entity.dt_updated.isoformat().replace(
        '+00:00', 'Z'
    )
    assert data['data']['attributes']['child_name'] == child_entity.child_name
    assert data['data']['attributes']['child_description'] == child_entity.child_description
    assert data['data']['attributes']['child_is_active'] == child_entity.child_is_active
    assert data['data']['attributes']['child_price'] == str(child_entity.child_price)
    assert data['data']['attributes'][
        'child_dt_approved'
    ] == child_entity.child_dt_approved.isoformat().replace('+00:00', 'Z')

    _relationships = data['data']['relationships']

    assert 'parent_entities' in _relationships

    _parent_entities = _relationships['parent_entities']

    assert len(_parent_entities['data']) == 1
    assert _parent_entities['data'][0]['id'] == str(parent_entity.id)
    assert _parent_entities['data'][0]['type'] == 'entity.parent_entity'

    # dependent_entity
    # LIST
    response = get_api_client(sample_app).get('/api/v1/entity/dependent_entity/')
    assert response.status_code == 200

    # RETRIEVE
    dependent_entity = DependentEntity.objects.first()
    response = get_api_client(sample_app).get(
        f'/api/v1/entity/dependent_entity/{str(dependent_entity.id)}/'
    )
    assert response.status_code == 200

    # extended_entity
    # LIST
    response = get_api_client(sample_app).get('/api/v1/entity/extended_entity/')
    assert response.status_code == 200

    # RETRIEVE
    extended_entity = ExtendedEntity.objects.first()
    response = get_api_client(sample_app).get(
        f'/api/v1/entity/extended_entity/{str(extended_entity.id)}/'
    )
    assert response.status_code == 200



# Copyright 2026 EcoFuture Technology Services LLC and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os
import sys

from django.conf import settings
from django.core.management.base import BaseCommand

import psutil


class Command(BaseCommand):
    """
    The command creates OpenAPI schemas for all supported languages
    by running a separate process for each language.

    Tags: RAG
    """

    def handle(self, **kwargs):
        envs_common = os.environ.copy()

        for lang_code, lang_name in settings.LANGUAGES:
            langs = json.dumps(
                sorted(settings.LANGUAGES, key=lambda x: x[0] != lang_code), ensure_ascii=False
            )

            envs = {**envs_common, **{'BS_LANGUAGE_CODE': lang_code, 'BS_LANGUAGES': langs}}

            print(f'Building schemas for {lang_name} ({lang_code})')

            psutil.Popen([sys.executable, 'manage.py', 'schemas_build_lang'], env=envs)

        psutil.wait_procs(psutil.Process().children())



#!/usr/bin/env python
# Copyright 2026 EcoFuture Technology Services LLC and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Django's command-line utility for administrative tasks."""
import os
import sys


def main():
    """Run administrative tasks."""
    os.environ.setdefault('DJANGO_SETTINGS_MODULE', 'settings')
    try:
        from django.core.management import execute_from_command_line
    except ImportError as exc:
        raise ImportError(
            "Couldn't import Django. Are you sure it's installed and "
            "available on your PYTHONPATH environment variable? Did you "
            "forget to activate a virtual environment?"
        ) from exc
    execute_from_command_line(sys.argv)


if __name__ == '__main__':
    main()



# Copyright 2026 EcoFuture Technology Services LLC and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from decimal import Decimal

import pytest
from bazis_test_utils.utils import get_api_client

from tests.utils.assert_sql import assert_sql_query, get_sql_query


@pytest.mark.django_db(transaction=True)
def test_related_one_relation_multiple_fields(sample_app, sample_vehicle_data):
    """
    @calc_property([FieldRelated('vehicle_model')])
    def vehicle_model_info(self) -> dict:
        return {
            'model': self.vehicle_model.model,
            'capacity': self.vehicle_model.capacity,
        }
    """

    vehicle = sample_vehicle_data['vehicle']
    vehicle_model = sample_vehicle_data['vehicle_model']

    response = get_api_client(sample_app).get(
        '/api/v1/related_one_relation_multiple_fields/entity/vehicle/'
    )
    assert response.status_code == 200

    sql_query = get_sql_query()
    expected_sql_template = """
        SELECT "entity_vehicle"."dt_created",
               "entity_vehicle"."dt_updated",
               "entity_vehicle"."id",
               "entity_vehicle"."vehicle_model_id",
               "entity_vehicle"."country_id",
               "entity_vehicle"."gnum",
               "entity_vehiclemodel"."dt_created",
               "entity_vehiclemodel"."dt_updated",
               "entity_vehiclemodel"."id",
               "entity_vehiclemodel"."brand_id",
               "entity_vehiclemodel"."model",
               "entity_vehiclemodel"."engine_type",
               "entity_vehiclemodel"."capacity"
        FROM "entity_vehicle"
        INNER JOIN "entity_vehiclemodel" ON ("entity_vehicle"."vehicle_model_id" = "entity_vehiclemodel"."id")
        LIMIT 20;"""
    assert_sql_query(expected_sql_template, sql_query)

    data = response.json()
    assert 'data' in data
    assert isinstance(data['data'], list)

    found = False
    for item in data['data']:
        if item['id'] != str(vehicle.id):
            continue

        found = True
        attrs = item['attributes']

        assert 'vehicle_model_info' in attrs
        assert attrs['vehicle_model_info'] == {
            'model': vehicle_model.model,
            'capacity': str(vehicle_model.capacity),
        }

    assert found, f'Vehicle with ID {vehicle.id} not found in response.'


@pytest.mark.django_db(transaction=True)
def test_related_hierarchy_list(sample_app, sample_vehicle_data):
    """
    @calc_property([FieldRelated('vehicle_model')])
    @calc_property(
        [FieldRelated('vehicle_model', alias='_vehicle_model',
            nested=[FieldRelated('brand', alias='_brand')])]
    )
    @calc_property([FieldRelated('vehicle_model__brand'), FieldRelated('country')])
    """

    vehicle = sample_vehicle_data['vehicle']
    vehicle_model = sample_vehicle_data['vehicle_model']
    vehicle_brand = sample_vehicle_data['vehicle_brand']
    vehicle_country = sample_vehicle_data['vehicle_country']

    response = get_api_client(sample_app).get('/api/v1/related/entity/vehicle/')
    assert response.status_code == 200

    sql_query = get_sql_query()
    expected_sql_template = """
        SELECT "entity_vehicle"."dt_created",
               "entity_vehicle"."dt_updated",
               "entity_vehicle"."id",
               "entity_vehicle"."vehicle_model_id",
               "entity_vehicle"."country_id",
               "entity_vehicle"."gnum",
               "entity_vehiclemodel"."dt_created",
               "entity_vehiclemodel"."dt_updated",
               "entity_vehiclemodel"."id",
               "entity_vehiclemodel"."brand_id",
               "entity_vehiclemodel"."model",
               "entity_vehiclemodel"."engine_type",
               "entity_vehiclemodel"."capacity",
               "entity_vehiclebrand"."dt_created",
               "entity_vehiclebrand"."dt_updated",
               "entity_vehiclebrand"."id",
               "entity_vehiclebrand"."name",
               "entity_country"."dt_created",
               "entity_country"."dt_updated",
               "entity_country"."id",
               "entity_country"."name"
        FROM "entity_vehicle"
        INNER JOIN "entity_vehiclemodel" ON ("entity_vehicle"."vehicle_model_id" = "entity_vehiclemodel"."id")
        INNER JOIN "entity_vehiclebrand" ON ("entity_vehiclemodel"."brand_id" = "entity_vehiclebrand"."id")
        INNER JOIN "entity_country" ON ("entity_vehicle"."country_id" = "entity_country"."id")
        LIMIT 20;"""
    assert_sql_query(expected_sql_template, sql_query)

    data = response.json()
    assert 'data' in data
    assert isinstance(data['data'], list)

    found = False
    for item in data['data']:
        if item['id'] != str(vehicle.id):
            continue

        found = True
        attrs = item['attributes']

        # FieldRelated - check vehicle_capacity 1
        assert 'vehicle_capacity_1' in attrs
        assert Decimal(attrs['vehicle_capacity_1']) == vehicle_model.capacity

        # FieldRelated - check vehicle_capacity 2
        assert 'vehicle_capacity_2' in attrs
        assert Decimal(attrs['vehicle_capacity_2']) == vehicle_model.capacity

        # FieldRelated - check vehicle_capacity 3
        assert 'vehicle_capacity_3' in attrs
        assert Decimal(attrs['vehicle_capacity_3']) == vehicle_model.capacity

        # FieldRelated - check brand multiple fields
        assert 'brand_info' in attrs
        assert attrs['brand_info'] == {
            'id': str(vehicle_brand.id),
            'name': vehicle_brand.name,
        }

        # FieldRelated - check brand and country
        assert 'brand_and_country' in attrs
        assert attrs['brand_and_country'] == {
            'brand': vehicle_brand.name,
            'country': vehicle_country.name,
        }

    assert found, f'Vehicle with ID {vehicle.id} not found in response.'


@pytest.mark.django_db(transaction=True)
def test_related_hierarchy_item(sample_app, sample_vehicle_data):
    """
    @calc_property([FieldRelated('vehicle_model')])
    @calc_property(
        [FieldRelated('vehicle_model', alias='_vehicle_model',
            nested=[FieldRelated('brand', alias='_brand')])]
    )
    @calc_property([FieldRelated('vehicle_model__brand'), FieldRelated('country')])
    """

    vehicle = sample_vehicle_data['vehicle']
    vehicle_model = sample_vehicle_data['vehicle_model']
    vehicle_brand = sample_vehicle_data['vehicle_brand']
    vehicle_country = sample_vehicle_data['vehicle_country']

    response = get_api_client(sample_app).get(f'/api/v1/related/entity/vehicle/{str(vehicle.id)}')

    assert response.status_code == 200

    sql_query = get_sql_query()
    expected_sql_template = """
        SELECT "entity_vehicle"."dt_created",
               "entity_vehicle"."dt_updated",
               "entity_vehicle"."id",
               "entity_vehicle"."vehicle_model_id",
               "entity_vehicle"."country_id",
               "entity_vehicle"."gnum",
               "entity_vehiclemodel"."dt_created",
               "entity_vehiclemodel"."dt_updated",
               "entity_vehiclemodel"."id",
               "entity_vehiclemodel"."brand_id",
               "entity_vehiclemodel"."model",
               "entity_vehiclemodel"."engine_type",
               "entity_vehiclemodel"."capacity",
               "entity_vehiclebrand"."dt_created",
               "entity_vehiclebrand"."dt_updated",
               "entity_vehiclebrand"."id",
               "entity_vehiclebrand"."name",
               "entity_country"."dt_created",
               "entity_country"."dt_updated",
               "entity_country"."id",
               "entity_country"."name"
        FROM "entity_vehicle"
        INNER JOIN "entity_vehiclemodel" ON ("entity_vehicle"."vehicle_model_id" = "entity_vehiclemodel"."id")
        INNER JOIN "entity_vehiclebrand" ON ("entity_vehiclemodel"."brand_id" = "entity_vehiclebrand"."id")
        INNER JOIN "entity_country" ON ("entity_vehicle"."country_id" = "entity_country"."id")
        WHERE "entity_vehicle"."id" = 'id_hex'::UUID
        ORDER BY "entity_vehicle"."id" ASC
        LIMIT 1;"""
    assert_sql_query(expected_sql_template, sql_query, vehicle.id.hex)

    data = response.json()
    attrs = data['data']['attributes']

    # FieldRelated - check vehicle_capacity 1
    assert 'vehicle_capacity_1' in attrs
    assert Decimal(attrs['vehicle_capacity_1']) == vehicle_model.capacity

    # FieldRelated - check vehicle_capacity 2
    assert 'vehicle_capacity_2' in attrs
    assert Decimal(attrs['vehicle_capacity_2']) == vehicle_model.capacity

    # FieldRelated - check vehicle_capacity 3
    assert 'vehicle_capacity_3' in attrs
    assert Decimal(attrs['vehicle_capacity_3']) == vehicle_model.capacity

    # FieldRelated - check brand multiple fields
    assert 'brand_info' in attrs
    assert attrs['brand_info'] == {
        'id': str(vehicle_brand.id),
        'name': vehicle_brand.name,
    }

    # FieldRelated - check brand and country multiple fields from several related tables
    assert 'brand_and_country' in attrs
    assert attrs['brand_and_country'] == {
        'brand': vehicle_brand.name,
        'country': vehicle_country.name,
    }



# Copyright 2026 EcoFuture Technology Services LLC and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import inspect
from collections.abc import Callable, Sequence
from enum import Enum, IntEnum
from importlib import import_module
from typing import Any

from fastapi import APIRouter, params
from fastapi._compat import (
    ModelField,
    annotation_is_pydantic_v1,
    lenient_issubclass,
)
from fastapi.datastructures import Default, DefaultPlaceholder
from fastapi.dependencies.utils import (
    _should_embed_body_fields,
    get_body_field,
    get_dependant,
    get_flat_dependant,
    get_parameterless_sub_dependant,
    get_typed_return_annotation,
)
from fastapi.exceptions import (
    PydanticV1NotSupportedError,
)
from fastapi.routing import APIRoute, APIWebSocketRoute, request_response
from fastapi.types import DecoratedCallable, IncEx
from fastapi.utils import (
    create_model_field,
    generate_unique_id,
    get_value_or_default,
    is_body_allowed_for_status_code,
)

from starlette import routing as starlette_routing
from starlette.responses import JSONResponse, Response
from starlette.routing import BaseRoute, compile_path, get_name


class BazisRoute(APIRoute):
    """
    Custom route class inheriting from FastAPI's APIRoute, with additional handling
    for response models, dependencies, and other route-specific configurations.
    """

    def __init__(
        self,
        path: str,
        endpoint: Callable[..., Any],
        *,
        response_model: Any = None,
        status_code: int | None = None,
        tags: list[str | Enum] | None = None,
        dependencies: Sequence[params.Depends] | None = None,
        summary: str | None = None,
        description: str | None = None,
        response_description: str = 'Successful Response',
        responses: dict[int | str, dict[str, Any]] | None = None,
        deprecated: bool | None = None,
        name: str | None = None,
        methods: set[str] | list[str] | None = None,
        operation_id: str | None = None,
        response_model_include: IncEx | None = None,
        response_model_exclude: IncEx | None = None,
        response_model_by_alias: bool = True,
        response_model_exclude_unset: bool = False,
        response_model_exclude_defaults: bool = False,
        response_model_exclude_none: bool = False,
        include_in_schema: bool = True,
        response_class: type[Response] | DefaultPlaceholder = Default(JSONResponse),
        dependency_overrides_provider: Any | None = None,
        callbacks: list[BaseRoute] | None = None,
        openapi_extra: dict[str, Any] | None = None,
        generate_unique_id_function: Callable[['APIRoute'], str] | DefaultPlaceholder = Default(
            generate_unique_id
        ),
    ) -> None:
        """
        Initializes a BazisRoute instance with the given parameters, setting up the
        path, endpoint, response model, dependencies, and other route-specific
        configurations.
        """
        self.path = path
        self.endpoint = endpoint
        if isinstance(response_model, DefaultPlaceholder):
            return_annotation = get_typed_return_annotation(endpoint)
            if lenient_issubclass(return_annotation, Response):
                response_model = None
            else:
                response_model = return_annotation
        self.response_model = response_model
        self.summary = summary
        self.response_description = response_description
        self.deprecated = deprecated
        self.operation_id = operation_id
        self.response_model_include = response_model_include
        self.response_model_exclude = response_model_exclude
        self.response_model_by_alias = response_model_by_alias
        self.response_model_exclude_unset = response_model_exclude_unset
        self.response_model_exclude_defaults = response_model_exclude_defaults
        self.response_model_exclude_none = response_model_exclude_none
        self.include_in_schema = include_in_schema
        self.response_class = response_class
        self.dependency_overrides_provider = dependency_overrides_provider
        self.callbacks = callbacks
        self.openapi_extra = openapi_extra
        self.generate_unique_id_function = generate_unique_id_function
        self.tags = tags or []
        self.responses = responses or {}
        self.name = get_name(endpoint) if name is None else name
        self.path_regex, self.path_format, self.param_convertors = compile_path(path)
        if methods is None:
            methods = ['GET']
        self.methods: set[str] = {method.upper() for method in methods}
        if isinstance(generate_unique_id_function, DefaultPlaceholder):
            current_generate_unique_id: Callable[[APIRoute], str] = (
                generate_unique_id_function.value
            )
        else:
            current_generate_unique_id = generate_unique_id_function
        self.unique_id = self.operation_id or current_generate_unique_id(self)
        # normalize enums e.g. http.HTTPStatus
        if isinstance(status_code, IntEnum):
            status_code = int(status_code)
        self.status_code = status_code
        if self.response_model:
            assert is_body_allowed_for_status_code(status_code), (
                f'Status code {status_code} must not have a response body'
            )
            response_name = 'Response_' + self.unique_id
            if annotation_is_pydantic_v1(self.response_model):
                raise PydanticV1NotSupportedError(
                    'pydantic.v1 models are no longer supported by FastAPI.'
                    f' Please update the response model {self.response_model!r}.'
                )
            self.response_field = create_model_field(
                name=response_name,
                type_=self.response_model,
                mode='serialization',
            )
            # Create a clone of the field, so that a Pydantic submodel is not returned
            # as is just because it's an instance of a subclass of a more limited class
            # e.g. UserInDB (containing hashed_password) could be a subclass of User
            # that doesn't have the hashed_password. But because it's a subclass, it
            # would pass the validation and be returned as is.
            # By being a new field, no inheritance will be passed as is. A new model
            # will be always created.
            # self.secure_cloned_response_field: Optional[
            #     ModelField
            # ] = create_cloned_field(self.response_field)
            self.secure_cloned_response_field: ModelField | None = self.response_field
        else:
            self.response_field = None  # type: ignore
            self.secure_cloned_response_field = None
        self.dependencies = list(dependencies or [])
        self.description = description or inspect.cleandoc(self.endpoint.__doc__ or '')
        # if a "form feed" character (page break) is found in the description text,
        # truncate description text to the content preceding the first "form feed"
        self.description = self.description.split('\f')[0].strip()
        response_fields = {}
        for additional_status_code, response in self.responses.items():
            assert isinstance(response, dict), 'An additional response must be a dict'
            model = response.get('model')
            if model:
                assert is_body_allowed_for_status_code(additional_status_code), (
                    f'Status code {additional_status_code} must not have a response body'
                )
                response_name = f'Response_{additional_status_code}_{self.unique_id}'
                if annotation_is_pydantic_v1(model):
                    raise PydanticV1NotSupportedError(
                        'pydantic.v1 models are no longer supported by FastAPI.'
                        f' In responses={{}}, please update {model}.'
                    )
                response_field = create_model_field(
                    name=response_name, type_=model, mode='serialization'
                )
                response_fields[additional_status_code] = response_field
        if response_fields:
            self.response_fields: dict[int | str, ModelField] = response_fields
        else:
            self.response_fields = {}

        assert callable(endpoint), 'An endpoint must be a callable'
        self.dependant = get_dependant(path=self.path_format, call=self.endpoint, scope='function')
        for depends in self.dependencies[::-1]:
            self.dependant.dependencies.insert(
                0,
                get_parameterless_sub_dependant(depends=depends, path=self.path_format),
            )
        self._flat_dependant = get_flat_dependant(self.dependant)
        self._embed_body_fields = _should_embed_body_fields(self._flat_dependant.body_params)
        self.body_field = get_body_field(
            flat_dependant=self._flat_dependant,
            name=self.unique_id,
            embed_body_fields=self._embed_body_fields,
        )
        self.app = request_response(self.get_route_handler())


class BazisDummyRoute(APIRoute):
    """
    A simplified version of BazisRoute, intended for use as a placeholder or default
    route class with basic configurations.
    """

    def __init__(
        self,
        path: str,
        endpoint: Callable[..., Any],
        *,
        response_model: Any = None,
        status_code: int | None = None,
        tags: list[str | Enum] | None = None,
        dependencies: Sequence[params.Depends] | None = None,
        summary: str | None = None,
        description: str | None = None,
        response_description: str = 'Successful Response',
        responses: dict[int | str, dict[str, Any]] | None = None,
        deprecated: bool | None = None,
        name: str | None = None,
        methods: set[str] | list[str] | None = None,
        operation_id: str | None = None,
        response_model_include: IncEx | None = None,
        response_model_exclude: IncEx | None = None,
        response_model_by_alias: bool = True,
        response_model_exclude_unset: bool = False,
        response_model_exclude_defaults: bool = False,
        response_model_exclude_none: bool = False,
        include_in_schema: bool = True,
        response_class: type[Response] | DefaultPlaceholder = Default(JSONResponse),
        dependency_overrides_provider: Any | None = None,
        callbacks: list[BaseRoute] | None = None,
        openapi_extra: dict[str, Any] | None = None,
        generate_unique_id_function: Callable[['APIRoute'], str] | DefaultPlaceholder = Default(
            generate_unique_id
        ),
    ) -> None:
        """
        Initializes a BazisDummyRoute instance with the given parameters, setting up the
        path, endpoint, response model, dependencies, and other basic configurations.
        """
        self.path = path
        self.endpoint = endpoint
        self.response_model = response_model
        self.summary = summary
        self.response_description = response_description
        self.deprecated = deprecated
        self.operation_id = operation_id
        self.response_model_include = response_model_include
        self.response_model_exclude = response_model_exclude
        self.response_model_by_alias = response_model_by_alias
        self.response_model_exclude_unset = response_model_exclude_unset
        self.response_model_exclude_defaults = response_model_exclude_defaults
        self.response_model_exclude_none = response_model_exclude_none
        self.include_in_schema = include_in_schema
        self.response_class = response_class
        self.dependency_overrides_provider = dependency_overrides_provider
        self.callbacks = callbacks
        self.openapi_extra = openapi_extra
        self.generate_unique_id_function = generate_unique_id_function
        self.tags = tags or []
        self.responses = responses or {}
        self.name = name
        if methods is None:
            methods = ['GET']
        self.methods: set[str] = {method.upper() for method in methods}
        if isinstance(status_code, IntEnum):
            status_code = int(status_code)
        self.status_code = status_code
        self.dependencies = list(dependencies or [])
        self.description = description


class BazisRouter(APIRouter):
    """
    Custom router class inheriting from FastAPI's APIRouter, with additional methods
    for route management and registration.

    Tags: RAG, EXPORT
    """

    def __init__(self, **kwargs) -> None:
        """
        Initializes a BazisRouter instance, setting the default route class to
        BazisDummyRoute and passing any additional keyword arguments to the superclass
        constructor.
        """
        kwargs.setdefault('route_class', BazisDummyRoute)
        super().__init__(**kwargs)

    def routes_cast(self, new_class: type[APIRoute] = APIRoute) -> list[APIRoute]:
        """
        Changes the route type to the specified one. This method is intended to work around
        the issue of the default route type setting in FastAPI, where the route type
        is determined through fastapi.routing.APIRouter.include_router:

        route_class_override=type(route)

        :param new_class: The class to which the route type should be changed.
        :return: A list of routes with the changed type.
        """
        for route in self.routes:
            route.__class__ = new_class
        return self.routes

    def register(self, prefix, arg=None, **kwargs):
        """
        Registers a new route or router with the given prefix and additional keyword
        arguments. Supports importing modules, including routers, and resetting routes.
        """
        if arg is None:
            arg = prefix
            prefix = ''

        if isinstance(arg, str):
            # example: router.register('entity.router')
            app_router = import_module(arg)
            self.include_router(app_router.router, prefix=prefix, **kwargs)
            if routers_with_prefix := getattr(app_router, 'routers_with_prefix', None):
                for router_prefix, router_value in routers_with_prefix.items():
                    self.include_router(router_value, prefix=f'/{router_prefix}', **kwargs)
        elif isinstance(arg, APIRouter):
            # example: router.register(routes.ChildEntityRouteSet.as_router())
            if hasattr(arg, 'get_url_prefix'):
                prefix = arg.get_url_prefix()
            self.include_router(arg, prefix=prefix, **kwargs)
        elif hasattr(arg, 'route'):
            self.reset_route(prefix, arg.route)

    def add_api_route(self, path: str, endpoint: Callable[..., Any], **kwargs):
        """
        Adds a new API route to the router, saving information about the installed route
        for the given endpoint.
        """
        super().add_api_route(path, endpoint, **kwargs)
        # save info about the installed route for this endpoint
        endpoint.route = self.routes[-1]

    def reset_route(self, path: str, route):
        """
        Resets an existing route with the specified path and route configurations,
        supporting various route types including APIRoute, starlette.routing.Route,
        APIWebSocketRoute, and starlette.routing.WebSocketRoute.
        """
        if isinstance(route, APIRoute):
            combined_responses = {**route.responses}
            use_response_class = get_value_or_default(
                route.response_class,
                self.default_response_class,
            )
            current_generate_unique_id = get_value_or_default(
                route.generate_unique_id_function,
                self.generate_unique_id_function,
            )
            self.add_api_route(
                path or route.path,
                route.endpoint,
                response_model=route.response_model,
                status_code=route.status_code,
                tags=route.tags.copy(),
                dependencies=route.dependencies.copy(),
                summary=route.summary,
                description=route.description,
                response_description=route.response_description,
                responses=combined_responses,
                deprecated=route.deprecated or self.deprecated,
                methods=route.methods,
                operation_id=route.operation_id,
                response_model_include=route.response_model_include,
                response_model_exclude=route.response_model_exclude,
                response_model_by_alias=route.response_model_by_alias,
                response_model_exclude_unset=route.response_model_exclude_unset,
                response_model_exclude_defaults=route.response_model_exclude_defaults,
                response_model_exclude_none=route.response_model_exclude_none,
                include_in_schema=route.include_in_schema and self.include_in_schema,
                response_class=use_response_class,
                name=route.name,
                route_class_override=type(route),
                callbacks=route.callbacks.copy(),
                openapi_extra=route.openapi_extra,
                generate_unique_id_function=current_generate_unique_id,
            )
        elif isinstance(route, starlette_routing.Route):
            methods = list(route.methods or [])  # type: ignore # in Starlette
            self.add_route(
                path or route.path,
                route.endpoint,
                methods=methods,
                include_in_schema=route.include_in_schema,
                name=route.name,
            )
        elif isinstance(route, APIWebSocketRoute):
            self.add_api_websocket_route(path or route.path, route.endpoint, name=route.name)
        elif isinstance(route, starlette_routing.WebSocketRoute):
            self.add_websocket_route(path or route.path, route.endpoint, name=route.name)

    def internal(
        self,
        path: str,
        *,
        response_model: type[Any] | None = None,
        status_code: int | None = None,
        tags: list[str | Enum] | None = None,
        dependencies: Sequence[params.Depends] | None = None,
        summary: str | None = None,
        description: str | None = None,
        response_description: str = 'Successful Response',
        responses: dict[int | str, dict[str, Any]] | None = None,
        deprecated: bool | None = None,
        operation_id: str | None = None,
        response_model_include: IncEx | None = None,
        response_model_exclude: IncEx | None = None,
        response_model_by_alias: bool = True,
        response_model_exclude_unset: bool = False,
        response_model_exclude_defaults: bool = False,
        response_model_exclude_none: bool = False,
        include_in_schema: bool = True,
        response_class: type[Response] = Default(JSONResponse),
        name: str | None = None,
        callbacks: list[BaseRoute] | None = None,
        openapi_extra: dict[str, Any] | None = None,
        generate_unique_id_function: Callable[[APIRoute], str] = Default(generate_unique_id),
    ) -> Callable[[DecoratedCallable], DecoratedCallable]:
        """
        Decorator method for defining internal routes with the specified path and
        additional configurations, including response model, status code, tags,
        dependencies, and more.
        """
        return self.api_route(
            path=path,
            response_model=response_model,
            status_code=status_code,
            tags=tags,
            dependencies=dependencies,
            summary=summary,
            description=description,
            response_description=response_description,
            responses=responses,
            deprecated=deprecated,
            methods=['INTERNAL'],
            operation_id=operation_id,
            response_model_include=response_model_include,
            response_model_exclude=response_model_exclude,
            response_model_by_alias=response_model_by_alias,
            response_model_exclude_unset=response_model_exclude_unset,
            response_model_exclude_defaults=response_model_exclude_defaults,
            response_model_exclude_none=response_model_exclude_none,
            include_in_schema=include_in_schema,
            response_class=response_class,
            name=name,
            callbacks=callbacks,
            openapi_extra=openapi_extra,
            generate_unique_id_function=generate_unique_id_function,
        )



# Copyright 2026 EcoFuture Technology Services LLC and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import dataclasses
from collections.abc import Callable, Sequence
from enum import Enum
from functools import partial
from typing import TYPE_CHECKING, Any

from fastapi.datastructures import Default
from fastapi.params import Depends as DependsCls
from fastapi.routing import APIRoute
from fastapi.types import IncEx
from fastapi.utils import generate_unique_id

from starlette.responses import JSONResponse, Response
from starlette.routing import BaseRoute

from pydantic import BaseModel

from bazis.core.schemas.enums import ApiAction


if TYPE_CHECKING:
    from .initial import InitialRouteBase


class EndpointType(Callable):
    """
    Callable type that represents an endpoint function. It includes a reference to
    the RouteContext.

    Tags: RAG
    """

    route_ctx: 'RouteContext'


@dataclasses.dataclass
class RouteParams:
    """
    Data class that contains parameters for a FastAPI route.

    Tags: RAG
    """

    path: str
    response_model: BaseModel | dict | None = None
    status_code: int | None = None
    tags: list[str | Enum] | None = (None,)
    dependencies: Sequence[DependsCls] | None = None
    summary: str | None = None
    description: str | None = (None,)
    response_description: str = 'Successful Response'
    responses: dict[int | str, dict[str, Any]] | None = None
    deprecated: bool | None = None
    operation_id: tuple = (None,)
    response_model_include: IncEx | None = None
    response_model_exclude: IncEx | None = None
    response_model_by_alias: bool = True
    response_model_exclude_unset: bool = False
    response_model_exclude_defaults: bool = False
    response_model_exclude_none: bool = False
    include_in_schema: bool = True
    response_class: type[Response] = dataclasses.field(
        default_factory=lambda: Default(JSONResponse)
    )
    name: str | None = None
    callbacks: list[BaseRoute] | None = None
    openapi_extra: dict[str, Any] | None = None
    generate_unique_id_function: Callable[[APIRoute], str] = dataclasses.field(
        default_factory=lambda: Default(generate_unique_id)
    )


@dataclasses.dataclass
class RouteContext:
    """
    An object of this class is a proxy wrapper for the route function.
    After a method becomes a route through one of the http\_... decorators,
    accessing it returns a RouteContext object.
    This object stores references to the original method, routing parameters,
    the endpoint function, and the route's local storage.
    Each inherited class will have its own RouteContext object.

    Tags: RAG
    """

    #: HTTP method associated with the route
    http_method: str
    #: Reference to the route parameters
    route_params: RouteParams
    #: Set of callbacks invoked during endpoint initialization
    endpoint_callbacks: list[Callable[..., Any] | partial] | None
    #: Tag to be injected into the route
    inject_tags: set[ApiAction] | None
    #: Function with the route logic
    func: Callable[..., Any] | None
    #: Route name
    name: str | None = None
    #: Reference to the router class
    route_cls: type['InitialRouteBase'] | None = None
    #: Final Inject class obtained by merging all Injects from parent classes
    Inject: type[dataclasses.dataclass] | None = None
    #: Function that will be passed as the route function
    endpoint: EndpointType | None = None
    #: Arbitrary route data storage
    store: dict = dataclasses.field(default_factory=dict)
    #: Reference to the fastapi route
    route: BaseRoute | None = None

    def __set_name__(self, owner: type['InitialRouteBase'], name: str):
        """
        If the magic method was called, the route function is a method, initialize the route in it.
        """
        self.route_cls = owner
        self.name = name
        self.route_cls.routes_ctx[self.name] = self



# Copyright 2026 EcoFuture Technology Services LLC and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys
from importlib import import_module
from pathlib import Path

from bazis.core.utils.imp import walk_packages_excluding


def get_apps_with_locals(package_name) -> set[str]:
    registered_packages = set()

    try:
        # Import the main package
        package = import_module(package_name)

        # If the package has no __path__ attribute, it is not a package
        if not hasattr(package, '__path__'):
            return registered_packages

        # Check whether the current package has a locale folder
        if package.__file__:
            package_path = os.path.dirname(package.__file__)
            locale_path = os.path.join(package_path, 'locale')

            if os.path.isdir(locale_path):
                # Add the package to the list of found ones
                registered_packages.add(package_name)
        # Recursively traverse all subpackages
        for _, name, is_pkg in walk_packages_excluding(
            package.__path__, package.__name__ + '.', exclude={'schemas', 'jsonapi'}
        ):
            if is_pkg:  # Process only packages, not modules
                # Recursive call for the subpackage
                subpackage_results = get_apps_with_locals(name)
                registered_packages.update(subpackage_results)

    except (ImportError, AttributeError) as e:
        print(f'Error while processing package {package_name}: {e}')

    return registered_packages


def discover_locale_paths(base_dir: str) -> list[str]:
    """
    Discovers and returns a list of paths to 'locale' directories within the given
    base directory and all imported packages. It first checks for a 'locale'
    directory in the base directory, then iterates through all imported packages to
    find their 'locale' directories.
    """
    locale_paths = []

    project_locale_path = os.path.join(base_dir, 'locale')
    if os.path.isdir(project_locale_path):
        locale_paths.append(project_locale_path)

    for package in set(sys.modules.keys()) | get_apps_with_locals('bazis'):
        try:
            module = import_module(package)
            package_path = Path(module.__file__).parent
            locale_path = package_path / 'locale'
            locale_path_str = str(locale_path)

            if locale_path.is_dir() and locale_path_str not in locale_paths:
                locale_paths.append(locale_path_str)
        except (ModuleNotFoundError, AttributeError, TypeError):
            continue

    return locale_paths



# Copyright 2026 EcoFuture Technology Services LLC and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Generated by Django 6.0.1 on 2026-01-14 12:22

import base64
import json
import pickle
from django.db import migrations


def is_pickle_data(value):
    """Checks if the value is pickle data in base64 format"""
    if not value:
        return False

    try:
        # Try to parse as JSON
        json.loads(value)
        return False  # Already JSON
    except (json.JSONDecodeError, TypeError, ValueError):
        pass

    try:
        # Try to decode as base64+pickle
        decoded = base64.b64decode(value.encode('ascii'))
        # Check if it looks like pickle (starts with pickle magic bytes)
        return decoded[:2] in [b'\x80\x02', b'\x80\x03', b'\x80\x04', b'\x80\x05']
    except Exception:
        return False


def convert_pickle_to_json(apps, schema_editor):
    """Converts constance data from pickle (base64) to JSON"""

    # Don't use apps.get_model, work directly with DB
    # because the model may change in the future
    db_alias = schema_editor.connection.alias
    cursor = schema_editor.connection.cursor()

    try:
        # Get all records
        cursor.execute('SELECT id, key, value FROM constance_constance')
        rows = cursor.fetchall()

        converted_count = 0
        skipped_count = 0
        error_count = 0

        for row_id, key, value in rows:
            try:
                # Check if conversion is needed
                if not is_pickle_data(value):
                    print(f'  Skipping {key} - not pickle or already JSON')
                    skipped_count += 1
                    continue

                # Decode base64
                try:
                    pickled_data = base64.b64decode(value.encode('ascii'))
                except Exception as e:
                    print(f'  ⚠ Base64 decode error for {key}: {e}')
                    error_count += 1
                    continue

                # Unpickle
                try:
                    unpickled_value = pickle.loads(pickled_data)
                except Exception as e:
                    print(f'  ⚠ Unpickle error for {key}: {e}')
                    error_count += 1
                    continue

                # Convert to JSON
                try:
                    json_value = json.dumps(unpickled_value, ensure_ascii=False)
                except (TypeError, ValueError) as e:
                    print(f'  ⚠ JSON serialization error for {key}: {e}')
                    error_count += 1
                    continue

                # Update in DB
                cursor.execute(
                    'UPDATE constance_constance SET value = %s WHERE id = %s',
                    [json_value, row_id]
                )

                print(
                    f'  ✓ Converted {key}: {type(unpickled_value).__name__} = {unpickled_value}'
                )
                converted_count += 1

            except Exception as e:
                print(f'  ✗ Unexpected error processing {key}: {e}')
                error_count += 1

        print(f'\n=== Conversion Summary ===')
        print(f'Converted: {converted_count}')
        print(f'Skipped: {skipped_count}')
        print(f'Errors: {error_count}')

        if error_count > 0:
            print(f'\n⚠ WARNING: There are errors! Check values manually.')

    except Exception as e:
        print(f'Critical migration error: {e}')
        raise


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0001_initial'),
    ]

    operations = [
        migrations.RunPython(
            convert_pickle_to_json,
            migrations.RunPython.noop,
        ),
    ]


# Copyright 2026 EcoFuture Technology Services LLC and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime

from fastapi.exceptions import ResponseValidationError

import pytest
from bazis_test_utils.utils import get_api_client


def truncate_to_seconds(dt: datetime) -> str | None:
    if dt is None:
        return None
    # Truncate microseconds and format in FastAPI style (with 'Z' instead of '+00:00')
    return dt.replace(microsecond=0).isoformat() + 'Z'  # 2025-04-30T08:44:02Z


@pytest.mark.django_db(transaction=True)
def test_get_vehicle_current_task(sample_app, route_injection_vehicle_data):
    """
    Checking the current transportation task for the vehicle.
    """

    ## Checking retrieval of the current transportation task for the vehicle.

    vehicle = route_injection_vehicle_data['vehicle']
    driver = route_injection_vehicle_data['driver']
    organization = route_injection_vehicle_data['organization']
    task2 = route_injection_vehicle_data['tasks'][1]

    vehicle_id = vehicle.id

    url = (
        f'/api/v1/custom_response_model/route_injection/vehicle/{vehicle_id}/current_carrier_task/'
    )

    response = get_api_client(sample_app).get(url)
    assert response.status_code == 200

    actual = response.json()

    expected = {
        'data': {
            'id': str(task2.id),
            'type': 'route_injection.carrier_task',
            'bs:action': 'view',
            'attributes': {
                'dt_created': truncate_to_seconds(task2.dt_created),
                'dt_updated': truncate_to_seconds(task2.dt_updated),
                'dt_start': truncate_to_seconds(task2.dt_start),
                'dt_finish': None,
                'fact_waste_weight': '4.500',
            },
            'relationships': {
                'vehicle': {'data': {'id': str(vehicle.id), 'type': 'route_injection.vehicle'}},
                'driver': {'data': {'id': str(driver.id), 'type': 'route_injection.driver'}},
                'org_owner': {
                    'data': {'id': str(organization.id), 'type': 'route_injection.organization'}
                },
            },
        }
    }

    # ===== Handling dt_created and dt_updated =====
    for field in ('dt_created', 'dt_updated', 'dt_start'):
        actual_value = actual['data']['attributes'].pop(field, None)
        expected['data']['attributes'].pop(field)

        assert isinstance(actual_value, str)
        # We do not compare the value directly, only check the type
        actual_dt = datetime.fromisoformat(actual_value.replace('Z', '+00:00'))
        assert isinstance(actual_dt, datetime)

    assert actual == expected

    ## Checking adjustment of the current transportation task for the vehicle.

    driver2 = route_injection_vehicle_data['driver2']

    response = get_api_client(sample_app).patch(
        url,
        json_data={
            'driver_id': str(driver2.id),
        },
    )

    assert response.status_code == 200

    actual = response.json()

    expected = {
        'data': {
            'id': str(task2.id),
            'type': 'route_injection.carrier_task',
            'bs:action': 'view',
            'attributes': {
                'dt_created': truncate_to_seconds(task2.dt_created),
                'dt_updated': truncate_to_seconds(task2.dt_updated),
                'dt_start': truncate_to_seconds(task2.dt_start),
                'dt_finish': None,
                'fact_waste_weight': '4.500',
            },
            'relationships': {
                'vehicle': {'data': {'id': str(vehicle.id), 'type': 'route_injection.vehicle'}},
                'driver': {'data': {'id': str(driver2.id), 'type': 'route_injection.driver'}},
                'org_owner': {
                    'data': {'id': str(organization.id), 'type': 'route_injection.organization'}
                },
            },
        },
        'meta': None,
    }

    # ===== Handling dt_created and dt_updated =====
    for field in ('dt_created', 'dt_updated', 'dt_start'):
        actual_value = actual['data']['attributes'].pop(field, None)
        expected['data']['attributes'].pop(field)

        assert isinstance(actual_value, str)
        # We do not compare the value directly, only check the type
        actual_dt = datetime.fromisoformat(actual_value.replace('Z', '+00:00'))
        assert isinstance(actual_dt, datetime)

    assert actual == expected


@pytest.mark.django_db(transaction=True)
def test_get_some_static(sample_app):
    """
    Checking custom response, positive scenario.

    For the endpoint method, it is explicitly specified that the response is a list of dictionaries, and therefore validation and response are without errors.

        def get_some_statics(self, **kwargs) -> list[dict]:
        # Check with correct response description
        return [{'k1': 'v1'}, {'k2': 'v2'}]
    """

    url = '/api/v1/custom_response_model/route_injection/vehicle/some_statics/'

    response = get_api_client(sample_app).get(url)
    assert response.status_code == 200

    actual = response.json()
    expected = [{'k1': 'v1'}, {'k2': 'v2'}]
    assert actual == expected


@pytest.mark.django_db(transaction=True)
def test_get_some_static_negative_1(sample_app):
    """
    Checking custom response, negative scenario.

    For the endpoint method, the response model is not explicitly specified, so the error only validates the presence of the id field.

        def get_some_statics_negative_1(self, **kwargs):
        # Reproducing the error when the response description is missing
        return [{'k1': 'v1'}, {'k2': 'v2'}]
    """

    url = '/api/v1/custom_response_model/route_injection/vehicle/some_statics_negative_1/'

    response = get_api_client(sample_app).get(url)
    assert response.status_code == 422

    actual = response.json()
    assert actual['errors'] == [
        {
            'code': 'ERR_VALIDATE',
            'detail': 'Field required',
            'source': {'pointer': '/id'},
            'status': 422,
            'title': 'missing',
        },
    ]


@pytest.mark.django_db(transaction=True)
def test_get_some_static_negative_2(sample_app):
    """
    Checking custom response, negative scenario.
    """

    url = '/api/v1/custom_response_model/route_injection/vehicle/some_statics_negative_2/'

    with pytest.raises(ResponseValidationError) as exc_info:
        response = get_api_client(sample_app).get(url)

    err = exc_info.value
    assert isinstance(err, ResponseValidationError)
    assert err.errors()[0]['type'] == 'dict_type'
    assert 'Input should be a valid dictionary' in err.errors()[0]['msg']

    from fastapi.testclient import TestClient

    from bazis.core.app import app

    client = TestClient(app, raise_server_exceptions=False)
    response = client.get(
        '/api/v1/custom_response_model/route_injection/vehicle/some_statics_negative_2/'
    )
    assert response.status_code == 500



# Copyright 2026 EcoFuture Technology Services LLC and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
# Demonstration of calculated fields in Django models

## Structure of `FieldCalc` descendants:
- `FieldRelated` — access to data of records referenced in the current table as FK (`select_related`).
- `FieldAnnotate`:
  - `FieldAggr` — applying SQL functions to fields of tables where the current model is an FK. Not used.
  - `FieldSubquery`:
    - `FieldSubAggr` — subquery with aggregate functions.
    - `FieldIsExists` — checks for the existence of records by conditions in the table where the current FK is.
- `FieldJson` — inherits from `FieldRelated` and `FieldSubquery`.

Common attributes:
    source: str
    query: Q | Expression = None
    alias: str = None
    context: list[str] = None
Related:
    slice: slice = None
    nested: list['FieldCalc'] = dataclasses.field(default_factory=list)
    fields: list[str] = dataclasses.field(default_factory=list)
    order_by: list[str] = None
    filter_fn: Callable[[QuerySet, dict], QuerySet] = None
Aggr
    func: Type[Func] = None
Subquery
    filter_fn: Callable[[QuerySet, dict], QuerySet] = None
SubAggr
    func: str = None
IsExists
    nested: list['FieldSubquery'] = dataclasses.field(default_factory=list)
Individual attributes:
    slice exists only for Json.
    order_by exists only for Json, because only it returns a list of elements.
    fields exist only for Json.
    nested exists for Related and Json and IsExists, and does not exist for Aggregates.
    func exists only for Aggr and SubAggr, Django aggregates or a string. It is recommended to use SubAggr.
    filter_fn exists for all except Aggr. TODO Cases.
Common attributes
    query - there are examples only for IsExists and SubAggr.
    context - there are examples for Json and Related.
    source
    alias - Required for Aggregates, IsExists and in the Related (Dynamic) hierarchy, in the json hierarchy it is auto-filled.
"""

from decimal import Decimal

from django.db import models
from django.db.models import F, Q, Sum

from bazis_test_utils.models_abstract import (
    ChildEntityBase,
    DependentEntityBase,
    ExtendedEntityBase,
    ParentEntityBase,
)

from bazis.core.fields import ChoiceArrayField
from bazis.core.models_abstract import DtMixin, JsonApiMixin, UuidMixin
from bazis.core.utils.functools import get_attr
from bazis.core.utils.orm import (
    FieldAggr,
    FieldIsExists,
    FieldJson,
    FieldRelated,
    FieldSubAggr,
    calc_property,
)


class Organization(DtMixin, UuidMixin, JsonApiMixin):
    """Division."""

    name = models.CharField(max_length=255)


class VehicleBrand(DtMixin, UuidMixin, JsonApiMixin):
    """Vehicle brand."""

    name = models.CharField('Brand Name', max_length=255, unique=True)

    class Meta:
        verbose_name = 'Vehicle Brand'
        verbose_name_plural = 'Vehicle Brands'

    def __str__(self):
        return self.name


class Country(DtMixin, UuidMixin, JsonApiMixin):
    """Country."""

    name = models.CharField('Country Name', max_length=255, unique=True)

    class Meta:
        verbose_name = 'Country'
        verbose_name_plural = 'Countries'

    def __str__(self):
        return self.name


class VehicleModel(DtMixin, UuidMixin, JsonApiMixin):
    """Vehicle model."""

    brand = models.ForeignKey(VehicleBrand, verbose_name='Brand', on_delete=models.CASCADE)
    model = models.CharField('Model', max_length=255, unique=True)
    engine_type = models.CharField('Engine Type', max_length=50, null=True, blank=True)
    capacity = models.DecimalField(
        'Capacity, t', max_digits=6, decimal_places=2, null=True, blank=True
    )

    class Meta:
        verbose_name = 'Vehicle Model'
        verbose_name_plural = 'Vehicle Models'
        unique_together = ('brand', 'model')

    def __str__(self):
        return self.model


class Division(DtMixin, UuidMixin, JsonApiMixin):
    """Division."""

    name = models.CharField(max_length=255)
    org_owner = models.ForeignKey(
        'Organization',
        blank=True,
        null=True,
        db_index=True,
        on_delete=models.SET_NULL,
    )

    @calc_property(
        [
            FieldJson(source='drivers', context=['_user'], query=Q(org_owner=F('_organization'))),
        ]
    )
    def drivers_list(self) -> list:
        return [
            {
                'id': driver['id'],
            }
            for driver in self._drivers
        ]

    @calc_property(
        [
            FieldJson('drivers1'),
        ]
    )
    def drivers_list1(self) -> list:
        return [
            {
                'id': driver['id'],
            }
            for driver in get_attr(self, '_drivers1') or []
        ]


class Driver(DtMixin, UuidMixin, JsonApiMixin):
    """Driver."""

    first_name = models.CharField('First Name', max_length=255)
    last_name = models.CharField('Last Name', max_length=255)
    contact_phone = models.CharField('Phone', max_length=50, null=True, blank=True)

    divisions = models.ManyToManyField(
        Division,
        related_name='drivers',
        blank=True,
    )
    org_owner = models.ForeignKey(
        'Organization',
        blank=True,
        null=True,
        db_index=True,
        on_delete=models.SET_NULL,
    )

    @calc_property(
        [
            FieldJson(
                source='carrier_tasks',
                fields=['id', 'dt_start', 'dt_finish', 'fact_waste_weight', 'vehicle_id'],
                nested=[
                    FieldJson(
                        source='vehicle',
                        fields=['gnum', 'dt_created'],
                        nested=[FieldJson(source='vehicle_model', fields=['engine_type'])],
                    )
                ],
            ),
        ]
    )
    def carrier_tasks_json_hierarchy(self) -> list:
        return [
            {
                'id': task['id'],
                'dt_start': task['dt_start'],
                'dt_finish': task['dt_finish'],
                'fact_waste_weight': task['fact_waste_weight'],
                'vehicle_gnum': get_attr(task, '_vehicle.0.gnum'),
            }
            for task in self._carrier_tasks
        ]

    @calc_property(
        [
            FieldJson(
                source='carrier_tasks',
                fields=['id', 'dt_start', 'dt_finish', 'fact_waste_weight', 'vehicle_id'],
                context=['_organization'],
                query=Q(org_owner=F('_organization')),
                nested=[
                    FieldJson(
                        source='vehicle',
                        fields=['gnum', 'dt_created'],
                        nested=[FieldJson(source='vehicle_model', fields=['engine_type'])],
                    )
                ],
            ),
        ]
    )
    def carrier_tasks_context(self) -> list:
        return [
            {
                'id': task['id'],
                'dt_start': task['dt_start'],
                'dt_finish': task['dt_finish'],
                'fact_waste_weight': task['fact_waste_weight'],
                'vehicle_gnum': get_attr(task, '_vehicle.0.gnum'),
            }
            for task in self._carrier_tasks
        ]

    @calc_property(
        [
            FieldJson(
                source='carrier_tasks',
                fields=['id', 'dt_start', 'dt_finish', 'fact_waste_weight'],
                query=Q(dt_finish__isnull=False),
                slice=slice(1),
                order_by=['-dt_finish'],
            ),
        ]
    )
    def last_trip(self) -> dict:
        return {
            'id': get_attr(self, '_carrier_tasks.0.id'),
            'dt_start': get_attr(self, '_carrier_tasks.0.dt_start'),
            'dt_finish': get_attr(self, '_carrier_tasks.0.dt_finish'),
            'fact_waste_weight': get_attr(self, '_carrier_tasks.0.fact_waste_weight'),
        }

    @calc_property(
        [
            FieldJson(
                source='carrier_tasks',
                fields=['id', 'dt_start', 'dt_finish', 'fact_waste_weight'],
                query=Q(dt_finish__isnull=False),
                order_by=['dt_start'],
            ),
        ]
    )
    def trips(self) -> list:
        return [
            {
                'id': trips['id'],
                'dt_start': trips['dt_start'],
                'dt_finish': trips['dt_finish'],
                'fact_waste_weight': trips['fact_waste_weight'],
            }
            for trips in self._carrier_tasks
        ]

    class Meta:
        verbose_name = 'Driver'
        verbose_name_plural = 'Drivers'

    def __str__(self):
        return f'{self.first_name} {self.last_name}'


class Vehicle(DtMixin, UuidMixin, JsonApiMixin):
    """Vehicle with calculated fields."""

    vehicle_model = models.ForeignKey(
        VehicleModel, verbose_name='Vehicle Model', on_delete=models.CASCADE
    )
    country = models.ForeignKey(Country, verbose_name='Country', on_delete=models.CASCADE)
    gnum = models.CharField('State Registration Number', max_length=50, unique=True)

    @calc_property([FieldRelated('vehicle_model')])
    def vehicle_capacity_1(self) -> Decimal:
        return get_attr(self, 'vehicle_model.capacity', Decimal(0.00))

    """
    FieldRelated
    data about the vehicle capacity from the table that is linked by a foreign key in the current model
    we return the value of a single field from the related table
    implementation option 2
    """

    @calc_property([FieldRelated('vehicle_model')])
    def vehicle_capacity_2(self) -> Decimal:
        return get_attr(self.vehicle_model, 'capacity', Decimal(0.00))

    """
    FieldRelated
    data about the vehicle capacity from the table that is linked by a foreign key in the current model
    we return the value of a single field from the related table
    implementation option 3
    """

    @calc_property([FieldRelated('vehicle_model')])
    def vehicle_capacity_3(self) -> Decimal:
        return self.vehicle_model.capacity

    """
    FieldRelated - data about the vehicle model name and model characteristics
    from the table that is linked by a foreign key in the current model
    we return a dictionary with data from several fields of the related table
    """

    @calc_property([FieldRelated('vehicle_model')])
    def vehicle_model_info(self) -> dict:
        return {
            'model': self.vehicle_model.model,
            'capacity': self.vehicle_model.capacity,
        }

    @calc_property(
        [
            FieldRelated(
                'vehicle_model',
                alias='_vehicle_model',
                nested=[FieldRelated('brand', alias='_brand')],
            )
        ]
    )
    def brand_info(self) -> dict:
        return {
            'id': self.vehicle_model.brand.id,
            'name': self.vehicle_model.brand.name,
        }

    # @calc_property([FieldRelated('vehicle_model__brand')])
    # def brand_info(self) -> dict:
    #     return {
    #         'id': self.vehicle_model.brand.id,
    #         'name': self.vehicle_model.brand.name,
    #     }

    @calc_property([FieldRelated('vehicle_model__brand'), FieldRelated('country')])
    def brand_and_country(self) -> dict:
        return {
            'brand': self.vehicle_model.brand.name,
            'country': self.country.name,
        }

    @calc_property(
        [
            FieldAggr(
                source='carrier_tasks__fact_waste_weight',
                func=Sum,
                alias='current_tasks_waste_weight_aggr',
                query=Q(carrier_tasks__dt_start__isnull=False)
                & Q(carrier_tasks__dt_finish__isnull=True),
            ),
        ],
        as_filter=True,
    )
    def current_tasks_waste_weight_aggr(self) -> Decimal | None:
        return get_attr(self, 'current_tasks_waste_weight_aggr', Decimal(0))

    @calc_property(
        [
            FieldSubAggr(
                source='carrier_tasks__fact_waste_weight',
                func='Sum',
                alias='current_tasks_waste_weight_subaggr',
                query=Q(dt_start__isnull=False) & Q(dt_finish__isnull=True),
            ),
        ]
    )
    def current_tasks_waste_weight_subaggr(self) -> Decimal | None:
        return get_attr(self, 'current_tasks_waste_weight_subaggr', Decimal(0))

    @calc_property(
        [
            FieldIsExists(
                source='carrier_tasks',
                query=Q(dt_start__isnull=False) & ~Q(dt_finish__isnull=False),
                alias='has_active_trip',
            ),
        ],
        as_filter=True,
    )
    def has_active_trip(self) -> bool:
        return get_attr(self, 'has_active_trip', False)

    @calc_property(
        [
            FieldIsExists(
                source='carrier_tasks',
                alias='has_active_trip_with_phone_join',
                query=Q(dt_start__isnull=False)
                & Q(dt_finish__isnull=True)
                & Q(driver__contact_phone__isnull=False)
                & ~Q(driver__contact_phone=''),
            )
        ]
    )
    def has_active_trip_with_phone_join(self) -> bool:
        return get_attr(self, 'has_active_trip_with_phone_join', False)

    @calc_property(
        [
            FieldIsExists(
                source='carrier_tasks',
                alias='has_active_trip_with_phone_subquery',
                query=Q(dt_start__isnull=False)
                & Q(dt_finish__isnull=True)
                & Q(_has_drivers_with_phone=True),
                nested=[
                    FieldIsExists(
                        source='driver',
                        alias='_has_drivers_with_phone',
                        query=Q(contact_phone__isnull=False) & ~Q(contact_phone=''),
                    )
                ],
            )
        ]
    )
    def has_active_trip_with_phone_subquery(self) -> bool:
        return get_attr(self, 'has_active_trip_with_phone_subquery', False)

    @calc_property(
        [
            FieldJson(
                source='carrier_tasks',
                fields=['id', 'dt_start', 'dt_finish', 'fact_waste_weight', 'driver_id'],
            ),
        ]
    )
    def carrier_task_list(self) -> list:
        return [
            {
                'id': task['id'],
                'dt_start': task['dt_start'],
                'dt_finish': task['dt_finish'],
                'fact_waste_weight': task['fact_waste_weight'],
                'driver_id': task['driver_id'],
            }
            for task in self._carrier_tasks
        ]

    class Meta:
        verbose_name = 'Vehicle'
        verbose_name_plural = 'Vehicles'

    def __str__(self):
        return self.gnum


class CarrierTask(DtMixin, UuidMixin, JsonApiMixin):
    """Register of transportations."""

    vehicle = models.ForeignKey(
        Vehicle, verbose_name='Vehicle', related_name='carrier_tasks', on_delete=models.CASCADE
    )
    driver = models.ForeignKey(
        Driver, verbose_name='Driver', related_name='carrier_tasks', on_delete=models.CASCADE
    )
    dt_start = models.DateTimeField('Start Time', null=True, blank=True)
    dt_finish = models.DateTimeField('Start Time', null=True, blank=True)
    fact_waste_weight = models.DecimalField(
        'Fact Waste Weight, t', max_digits=15, decimal_places=3, null=True, blank=True
    )
    org_owner = models.ForeignKey(
        'Organization',
        blank=True,
        null=True,
        db_index=True,
        on_delete=models.SET_NULL,
    )

    class Meta:
        verbose_name = 'Carrier Task'
        verbose_name_plural = 'Carrier Tasks'
        indexes = [
            models.Index(fields=['vehicle', 'dt_start', 'dt_finish'], name='ix_vehicle_dates'),
        ]

    def __str__(self):
        return f'Carrier Task {self.id}'


class ChildEntity(ChildEntityBase, DtMixin, UuidMixin, JsonApiMixin):
    pass


class DependentEntity(DependentEntityBase, DtMixin, UuidMixin, JsonApiMixin):
    parent_entity = models.ForeignKey(
        'ParentEntity', on_delete=models.CASCADE, related_name='dependent_entities'
    )


class DependentEntityNull(DependentEntityBase, DtMixin, UuidMixin, JsonApiMixin):
    parent_entity = models.ForeignKey(
        'ParentEntity', on_delete=models.CASCADE, related_name='dependent_entities_null', null=True
    )


class ExtendedEntity(ExtendedEntityBase, DtMixin, UuidMixin, JsonApiMixin):
    parent_entity = models.OneToOneField(
        'ParentEntity', on_delete=models.CASCADE, related_name='extended_entity'
    )


class ExtendedEntityNull(ExtendedEntityBase, DtMixin, UuidMixin, JsonApiMixin):
    parent_entity = models.OneToOneField(
        'ParentEntity', on_delete=models.CASCADE, related_name='extended_entity_null', null=True
    )


class ParentEntityState(models.TextChoices):
    STATE_ONE = (
        'state_one',
        'State one',
    )
    STATE_TWO = (
        'state_two',
        'State two',
    )
    STATE_THREE = (
        'state_three',
        'State three',
    )
    STATE_FOUR = (
        'state_four',
        'State four',
    )


class ParentEntityField(models.TextChoices):
    FIRST_FIELD = (
        'first_field',
        'First field',
    )
    SECOND_FIELD = (
        'second_field',
        'Second field',
    )
    THIRD_FIELD = (
        'third_field',
        'Third field',
    )


class ParentEntity(ParentEntityBase, DtMixin, UuidMixin, JsonApiMixin):
    child_entities = models.ManyToManyField(
        ChildEntity,
        related_name='parent_entities',
        blank=True,
    )
    state = models.CharField(
        'State',
        max_length=255,
        choices=ParentEntityState,
        default=ParentEntityState.STATE_ONE.value,
    )
    field = ChoiceArrayField(
        models.CharField(choices=ParentEntityField, max_length=32),
        verbose_name='Field',
        blank=True,
        null=True,
    )

    @calc_property([FieldRelated('extended_entity')])
    def extended_entity_price(self) -> Decimal:
        return get_attr(self, 'extended_entity.extended_price', Decimal(0.00))

    @calc_property(
        [
            FieldJson(
                source='child_entities',
                query=Q(child_is_active=True),
                fields=['id', 'child_name', 'child_description', 'child_is_active'],
            ),
        ]
    )
    def active_children(self) -> list:
        return [
            {
                'id': child['id'],
                'child_name': child['child_name'],
                'child_description': child['child_description'],
                'child_is_active': child['child_is_active'],
            }
            for child in self._child_entities
        ]

    @calc_property(
        [
            FieldSubAggr(
                'child_entities',
                query=Q(child_is_active=True),
                func='Count',
                alias='count_active_children',
            ),
        ]
    )
    def count_active_children(self) -> int:
        return self.count_active_children

    @calc_property(
        [
            FieldIsExists(
                'child_entities',
                query=Q(child_price__isnull=True) | Q(child_is_active=False),
                alias='has_inactive_children',
            ),
        ],
        as_filter=True,
    )
    def has_inactive_children(self) -> bool:
        return getattr(self, 'has_inactive_children', False)


class WithProtectedEntity(DtMixin, UuidMixin, JsonApiMixin):
    name = models.CharField(max_length=255)
    child = models.ForeignKey(
        ChildEntity,
        on_delete=models.PROTECT,
    )

    class Meta:
        verbose_name = 'Entity with Protected'
        verbose_name_plural = 'Entities with Protected'


class WithProtectedEntitySystem(DtMixin, UuidMixin):
    name = models.CharField(max_length=255)
    child = models.ForeignKey(
        ChildEntity,
        on_delete=models.PROTECT,
    )

    class Meta:
        verbose_name = 'Entity with Protected System'
        verbose_name_plural = 'Entities with Protected System'



# Copyright 2026 EcoFuture Technology Services LLC and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections.abc
from weakref import proxy


class Link:
    """
    Represents a node in a doubly linked list. It contains references to the
    previous and next nodes, the stored key, and a weak reference to itself.
    """

    __slots__ = 'prev', 'next', 'key', '__weakref__'


class OrderedSet(collections.abc.MutableSet):
    """
    Set that remembers the order elements were added. Provides O(1) time complexity
    for add, remove, and check operations, similar to regular sets.
    """

    # Big-O running times for all methods are the same as for regular sets.
    # The internal self.__map dictionary maps keys to links in a doubly linked list.
    # The circular doubly linked list starts and ends with a sentinel element.
    # The sentinel element never gets deleted (this simplifies the algorithm).
    # The prev/next links are weakref proxies (to prevent circular references).
    # Individual links are kept alive by the hard reference in self.__map.
    # Those hard references disappear when a key is deleted from an OrderedSet.

    def __init__(self, iterable=None):
        """
        Initializes an empty OrderedSet. If an iterable is provided, adds its elements
        to the set.
        """
        self.__root = root = Link()  # sentinel node for doubly linked list
        root.prev = root.next = root
        self.__map = {}  # key --> link
        if iterable is not None:
            self |= iterable

    def __len__(self):
        """
        Returns the number of elements in the OrderedSet.
        """
        return len(self.__map)

    def __contains__(self, key):
        """
        Checks if a key is in the OrderedSet.
        """
        return key in self.__map

    def add(self, key):
        """
        Adds a key to the OrderedSet, storing it in a new link at the end of the linked
        list.
        """
        if key not in self.__map:
            self.__map[key] = link = Link()
            root = self.__root
            last = root.prev
            link.prev, link.next, link.key = last, root, key
            last.next = root.prev = proxy(link)

    def discard(self, key):
        # Remove an existing item using self.__map to find the link which is
        """
        Removes an existing item using self.__map to find the link, which is then
        removed by updating the links in the predecessor and successor.
        """
        if key in self.__map:
            link = self.__map.pop(key)
            link.prev.next = link.next
            link.next.prev = link.prev

    def __iter__(self):
        """
        Traverses the linked list in order, yielding each key.
        """
        root = self.__root
        curr = root.next
        while curr is not root:
            yield curr.key
            curr = curr.next

    def __reversed__(self):
        """
        Traverses the linked list in reverse order, yielding each key.
        """
        root = self.__root
        curr = root.prev
        while curr is not root:
            yield curr.key
            curr = curr.prev

    def pop(self, last=True):
        """
        Removes and returns the last item if 'last' is True, otherwise removes and
        returns the first item. Raises KeyError if the set is empty.
        """
        if not self:
            raise KeyError('set is empty')
        key = next(reversed(self)) if last else next(iter(self))
        self.discard(key)
        return key

    def __repr__(self):
        """
        Returns a string representation of the OrderedSet.
        """
        if not self:
            return f'{self.__class__.__name__}()'
        return f'{self.__class__.__name__}({list(self)!r})'

    def __eq__(self, other):
        """
        Checks if two OrderedSets are equal by comparing their lengths and elements, or
        checks if they are disjoint if the other object is not an OrderedSet.
        """
        if isinstance(other, OrderedSet):
            return len(self) == len(other) and list(self) == list(other)
        return not self.isdisjoint(other)



# Copyright 2026 EcoFuture Technology Services LLC and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from django.conf import settings
from django.contrib import admin
from django.urls import path, re_path
from django.views.static import serve


urlpatterns = [
    path('admin/', admin.site.urls),
]

if settings.DEBUG:
    urlpatterns += [
        re_path(r'^media/(?P<path>.*)$', serve, dict(document_root=settings.MEDIA_ROOT)),
        re_path(r'^static/(?P<path>.*)$', serve, dict(document_root=settings.STATIC_ROOT)),
    ]



# Copyright 2026 EcoFuture Technology Services LLC and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Generated by Django 5.0.14 on 2025-12-24 15:29

import django.db.models.deletion
import model_clone.mixin
import uuid
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('sparse_fieldsets', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='Category',
            fields=[
                ('dt_created', models.DateTimeField(auto_now_add=True, db_index=True, verbose_name='Creation time')),
                ('dt_updated', models.DateTimeField(auto_now=True, db_index=True, verbose_name='Update time')),
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('name', models.CharField(max_length=255)),
                ('comment', models.CharField(max_length=255)),
            ],
            options={
                'abstract': False,
            },
            bases=(model_clone.mixin.CloneMixin, models.Model),
        ),
        migrations.AddField(
            model_name='article',
            name='category',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='articles', to='sparse_fieldsets.category'),
        ),
    ]



# Copyright 2026 EcoFuture Technology Services LLC and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import hashlib

from django.db import models

import pghistory
from pgtrigger.core import MAX_NAME_LENGTH


def trigger_name(name: str):
    """
    Generate a unique trigger name that complies with the maximum length constraint
    by truncating and appending a hash if necessary.

    Tags: RAG, EXPORT
    """
    if len(name) > MAX_NAME_LENGTH - 6:
        name = name[: MAX_NAME_LENGTH - 6] + hashlib.md5(name.encode()).hexdigest()[:6]
    return name


def register(*triggers, with_through=False):
    """
    Decorator to register triggers on a Django model, with an optional parameter to
    include through models for many-to-many relationships.

    Tags: RAG, EXPORT
    """

    def _model_wrapper(model_class):
        """
        Inner wrapper function for the register decorator that applies the trigger
        registration logic to a given model class.
        """

        def triggers_set(model):
            """
            Set triggers on the model, including through models for many-to-many
            relationships if specified.
            """
            if not model._meta.abstract and not model._meta.proxy:
                for trigger in triggers:
                    trigger._primary_model = model
                    trigger.register(model)

                    if with_through:
                        for m2m_field in model._meta.many_to_many:
                            through = getattr(model, m2m_field.name).through
                            if through._meta.auto_created:
                                trigger.register(through)

        def finalize(sender, **kwargs):
            """
            Signal handler to finalize trigger registration when the model class is
            prepared.
            """
            if not issubclass(sender, model_class):
                return
            triggers_set(sender)

        if model_class._meta.abstract:
            models.signals.class_prepared.connect(finalize, weak=False)
        else:
            triggers_set(model_class)

        return model_class

    return _model_wrapper


def history_track(*args, **kwargs):
    """
    Decorator to track history events on a Django model using pghistory, with
    support for many-to-many relationships.

    Tags: RAG, EXPORT
    """
    kwargs = {**{'context_field': pghistory.ContextJSONField()}, **kwargs}

    def _model_wrapper(model_class):
        """
        Inner wrapper function for the history_track decorator that applies the history
        tracking logic to a given model class.
        """

        def history_set(model):
            """
            Set history tracking on the model, including through models for many-to-many
            relationships if they are auto-created.
            """
            if not model._meta.abstract and not model._meta.proxy:
                pghistory.track(*args, **kwargs)(model)

                for m2m_field in model._meta.many_to_many:
                    through = getattr(model, m2m_field.name).through
                    if through._meta.auto_created:
                        pghistory.track(
                            pghistory.InsertEvent('add'),
                            pghistory.DeleteEvent('remove'),
                            **{**kwargs, **{'obj_field': None}},
                        )(through)

        def finalize(sender, **kwargs):
            """
            Signal handler to finalize history tracking when the model class is prepared.
            """
            if not issubclass(sender, model_class):
                return
            history_set(sender)

        if model_class._meta.abstract:
            models.signals.class_prepared.connect(finalize, weak=False)
        else:
            history_set(model_class)

        return model_class

    return _model_wrapper



# Copyright 2026 EcoFuture Technology Services LLC and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json

from pghistory import config, runtime


def _inject_history_context(execute, sql, params, many, context):
    """
    Eliminates the possibility of SQL exploit

    Tags: RAG, EXPORT
    """
    cursor = context['cursor']

    if not cursor.name and not runtime._is_concurrent_statement(sql):
        metadata_str = json.dumps(
            runtime._tracker.value.metadata, cls=config.json_encoder()
        ).replace("'", "''")

        params = (str(runtime._tracker.value.id), metadata_str) + (
            tuple(params) if params else tuple()
        )

        sql = ('SET LOCAL pghistory.context_id=%s;SET LOCAL pghistory.context_metadata=%s;') + sql

    return execute(sql, params, many, context)


runtime._inject_history_context = _inject_history_context



# Copyright 2026 EcoFuture Technology Services LLC and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum
from typing import Literal
from uuid import UUID

from django.db import models, transaction

from fastapi import HTTPException

from bazis.core.models_abstract import JsonApiMixin
from bazis.core.routes_abstract.jsonapi.schemas import ResourceIdentifier


class RelationType(Enum):
    M2M = 'm2m'
    O2M = 'o2m'
    M2O = 'm2o'
    O2O = 'o2o'


class RelationshipsService:
    """
    Service for managing JSON:API relationships.

    Tags: RAG, INTERNAL
    """

    @staticmethod
    def get_relation_type(rel_field) -> RelationType:
        if getattr(rel_field, 'many_to_many', False):
            return RelationType.M2M
        if getattr(rel_field, 'one_to_many', False):
            return RelationType.O2M
        if getattr(rel_field, 'many_to_one', False):
            return RelationType.M2O
        if getattr(rel_field, 'one_to_one', False) or isinstance(rel_field, models.OneToOneRel):
            return RelationType.O2O
        raise HTTPException(status_code=400, detail='Unknown relation type')

    @staticmethod
    def get_relationships_managment_set(
        model: type[JsonApiMixin], item_id: str, related_field_name: str
    ) -> tuple:
        item = model.objects.get(pk=item_id)
        rel_field = model._meta.get_field(related_field_name)
        manager = getattr(item, related_field_name, None)
        return item, rel_field, manager

    @staticmethod
    def _parse_id(raw_id: str, model: type[models.Model]) -> UUID | int | str:
        pk_field = model._meta.pk
        try:
            if isinstance(pk_field, models.UUIDField):
                return UUID(str(raw_id))
            elif isinstance(pk_field, models.IntegerField):
                return int(raw_id)
            return str(raw_id)
        except (ValueError, TypeError):
            raise HTTPException(
                400, f"Invalid id format '{raw_id}' for model '{model.__name__}'"
            ) from None

    @staticmethod
    def parse_ids(rel_field, relationships_data, to_many: bool) -> list:
        data = relationships_data.data
        if to_many:
            if not isinstance(data, list):
                raise HTTPException(400, f"To-Many field '{rel_field.name}' expects an array of objects")
        else:
            if not isinstance(data, (ResourceIdentifier, dict)):
                raise HTTPException(409, f"To-One field '{rel_field.name}' accepts a single object")

        related_model = rel_field.related_model
        ids = []

        if to_many:
            for row in data:
                obj_id = row.id if isinstance(row, ResourceIdentifier) else row['id']
                ids.append(RelationshipsService._parse_id(obj_id, related_model))
        else:
            obj_id = data.id if isinstance(data, ResourceIdentifier) else data['id']
            ids.append(RelationshipsService._parse_id(obj_id, related_model))

        existing_ids = set(related_model.objects.filter(id__in=ids).values_list('id', flat=True))
        missing_ids = set(ids) - existing_ids
        if missing_ids:
            raise HTTPException(
                400,
                f"Objects of model '{related_model.__name__}' with id {list(missing_ids)} not found",
            )
        return list(existing_ids)

    @staticmethod
    def parse_targets(rel_field, relationships_data, to_many: bool):
        ids = RelationshipsService.parse_ids(rel_field, relationships_data, to_many)
        model = rel_field.related_model
        return model.objects.filter(pk__in=ids) if to_many else model.objects.get(pk=ids[0])

    @staticmethod
    @transaction.atomic
    def update_o2o(item, rel_field, new_item, action: str):
        """
        Safely updates a OneToOne relationship.

        Args:
            item: Main object
            rel_field: Relation field
            new_item: New related object (None for remove)
            action: Action ("set", "add", "remove")
        """
        if action == 'remove':
            # Clear the relation
            if isinstance(rel_field, models.OneToOneRel):
                # Reverse relation - find and clear the related object
                related_name = rel_field.field.name
                current_related = getattr(item, rel_field.get_accessor_name(), None)
                if current_related:
                    setattr(current_related, related_name, None)
                    current_related.save(update_fields=[related_name])
            else:
                # Direct relation - clear it on the main object
                setattr(item, rel_field.name, None)
                item.save(update_fields=[rel_field.name])
            return

        # Set a new relation
        related_model = rel_field.related_model

        # Type check for the new object
        if new_item and not isinstance(new_item, related_model):
            raise HTTPException(
                400,
                f"Cannot assign an object of type '{type(new_item).__name__}' to field '{rel_field.name}': "
                f'expected an instance of {related_model.__name__}.',
            )

        if isinstance(rel_field, models.OneToOneRel):
            related_name = rel_field.field.name

            # Remove the current relation from the new object (if any)
            if hasattr(new_item, related_name) and getattr(new_item, related_name):
                current_owner = getattr(new_item, related_name)
                if current_owner != item:
                    setattr(new_item, related_name, None)
                    new_item.save(update_fields=[related_name])

            # Remove the current relation from the main object
            try:
                current_related = getattr(item, rel_field.get_accessor_name(), None)
                if current_related and current_related != new_item:
                    setattr(current_related, related_name, None)
                    current_related.save(update_fields=[related_name])
            except related_model.DoesNotExist:
                pass  # There was no relation - do nothing

            # Set a new relation
            setattr(new_item, related_name, item)
            new_item.save(update_fields=[related_name])

        else:
            # Direct relation
            related_name = rel_field.name

            # Remove the current relation from the new object (if any)
            if hasattr(new_item, related_name) and getattr(new_item, related_name):
                current_owner = getattr(new_item, related_name)
                if current_owner != item:
                    setattr(new_item, related_name, None)
                    new_item.save(update_fields=[related_name])

            # Set a new relation
            setattr(item, related_name, new_item)
            item.save(update_fields=[related_name])

    # Universal method
    @staticmethod
    def apply_relationship_action(
        action: Literal['add', 'set', 'remove'],
        model: type[JsonApiMixin],
        item_id: str,
        related_field_name: str,
        relationships_data,
    ) -> None:
        item, rel_field, manager = RelationshipsService.get_relationships_managment_set(
            model, item_id, related_field_name
        )
        relation_type = RelationshipsService.get_relation_type(rel_field)

        if relation_type in (RelationType.M2M, RelationType.O2M):
            targets = RelationshipsService.parse_targets(
                rel_field, relationships_data, to_many=True
            )
            if action == 'add':
                manager.add(*targets)
            elif action == 'remove':
                manager.remove(*targets)
            elif action == 'set':
                manager.set(targets)

        elif relation_type in (RelationType.O2O, RelationType.M2O):
            target = (
                RelationshipsService.parse_targets(rel_field, relationships_data, to_many=False)
                if relationships_data.data and action != 'remove'
                else None
            )
            RelationshipsService.update_o2o(item, rel_field, target, action)



# Copyright 2026 EcoFuture Technology Services LLC and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .cache import with_cache_openapi_schema
from .callbacks import (
    JsonApiResponse,
    api_action_init,
    api_action_jsonapi_init,
    api_action_response_init,
    item_data_typing,
    item_id_typing,
    meta_fields_addition,
)
from .mixins import DtRouteMixin, RestrictedQsRouteMixin, UniqNumberRouteMixin
from .route_base import JsonapiRouteBase



# Copyright 2026 EcoFuture Technology Services LLC and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging


logger = logging.getLogger(__name__)


def force_global_logging_level(target_level=logging.INFO, block_higher_levels=True):
    """
    Forcibly sets logging level for ALL loggers in the system

    Args:
        target_level: Target level (logging.DEBUG, INFO, WARNING, ERROR, CRITICAL or string)
        block_higher_levels: Whether to block attempts to set higher levels

    Tags: RAG, EXPORT
    """

    # Convert string to number if needed
    if isinstance(target_level, str):
        target_level = getattr(logging, target_level.upper(), logging.INFO)

    level_name = logging.getLevelName(target_level)

    def apply_level_to_all():
        # 1. Root logger
        logging.getLogger().setLevel(target_level)

        # 2. ALL existing named loggers
        count = 0
        for name in list(logging.Logger.manager.loggerDict.keys()):
            logger = logging.getLogger(name)
            logger.setLevel(target_level)
            count += 1

        return count

    # Apply to all existing loggers
    existing_count = apply_level_to_all()

    # 3. INTERCEPT creation of new loggers
    original_getLogger = logging.getLogger  # noqa: N806

    def patched_getLogger(name=None):  # noqa: N802
        logger = original_getLogger(name)
        logger.setLevel(target_level)  # Force our level for ALL new loggers
        return logger

    logging.getLogger = patched_getLogger

    # 4. OPTIONALLY block setting higher levels
    if block_higher_levels:
        original_setLevel = logging.Logger.setLevel  # noqa: N806

        def controlled_setLevel(self, level):  # noqa: N802
            # Convert to number if string
            if isinstance(level, str):
                numeric_level = getattr(logging, level.upper(), target_level)
            else:
                numeric_level = level

            # If trying to set level higher than target - block it
            if numeric_level > target_level:
                logger.debug(
                    f"🚫 BLOCKING setLevel({logging.getLevelName(numeric_level)}) for '{self.name}', keeping {level_name}"
                )
                numeric_level = target_level

            return original_setLevel(self, numeric_level)

        logging.Logger.setLevel = controlled_setLevel
        block_msg = f' with blocking levels above {level_name}'
    else:
        block_msg = ''

    logger.debug(f'✅ SET {level_name} for ALL {existing_count + 1} loggers{block_msg}')



# Copyright 2026 EcoFuture Technology Services LLC and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from urllib.parse import urlencode

import pytest
from bazis_test_utils.utils import get_api_client
from entity.models import (
    ChildEntity,
    ParentEntity,
)

from tests import factories


@pytest.mark.django_db(transaction=True)
def test_routes_searching(sample_app):
    searching_phrases = [
        'Simple text for searching',
        'Another simple text for searching',
        'This is a simple text for searching',
        'Complex text for searching',
        'Another complex text for searching',
        'This is a complex text for searching',
    ]

    for phrase in searching_phrases:
        factories.ParentEntityFactory.create(description=phrase, child_entities=True)

    query = urlencode(
        {
            'filter': 'description__$search=complex|description__$search=this',
        }
    )

    response = get_api_client(sample_app).get(f'/api/v1/entity/parent_entity/?{query}')

    assert response.status_code == 200

    data = response.json()

    for it in data['data']:
        description = it['attributes']['description'].lower()
        assert 'complex' in description or 'this' in description

    child = factories.ChildEntityFactory.create(child_description='Child text')
    parent = ParentEntity.objects.first()
    parent.child_entities.add(child)

    query = urlencode(
        {
            'filter': 'child_entities__child_description__$search=child',
        }
    )

    response = get_api_client(sample_app).get(f'/api/v1/entity/parent_entity/?{query}')

    assert response.status_code == 200

    data = response.json()

    for it in data['data']:
        assert ChildEntity.objects.filter(
            parent_entities=it['id'],
            child_description__icontains='child',
        ).exists()



# Copyright 2026 EcoFuture Technology Services LLC and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import ast
import importlib.util
import inspect
from pathlib import Path

from sphinx.ext.autodoc import FunctionDocumenter


def generate_dynamic_fields_rst(srcdir: str):
    dynamic_tests_root = Path(srcdir).parent / "tests" / "test_routes_view" / "dynamic"
    dynamic_fields_rst = Path(srcdir) / "dynamic_fields.rst"

    content = [
        "Dynamic fields",
        "==============",
        "",
        """
``FieldDynamic`` is a declarative builder for computed fields, used through the ``@calc_property`` decorator.

When calling ``FieldDynamic()``, it automatically transforms into one of the ``Field*`` classes:

- ``FieldIsExists`` (boolean indicator for a specified combination of conditions) — if ``alias`` starts with ``has_``.
- ``FieldSubAggr`` (aggregation result) — if ``func`` is specified.
- ``FieldJson`` (list of values for one-to-many and many-to-many relationships) — if ``fields`` are defined.
- ``FieldRelated`` (values for one-to-one and many-to-one relationships) — in all other cases.

``DependsCalc`` is a helper mechanism for passing dependencies.
It creates a Pydantic model from fields specified in ``FieldDynamic``, taking nesting into account.
Used in ``@calc_property`` to pass dependency values to methods.
        """,
        ""
    ]

    for py_file in sorted(dynamic_tests_root.rglob("test_*.py")):
        rel_path = py_file.relative_to(Path(srcdir).parent).with_suffix("")
        module = ".".join(rel_path.parts)

        spec = importlib.util.spec_from_file_location(module, str(py_file))
        mod = importlib.util.module_from_spec(spec)
        try:
            spec.loader.exec_module(mod)
            doc = inspect.getdoc(mod)
            if not doc:
                continue
            module_title = doc.strip().splitlines()[0]
        except Exception:
            continue

        content.append("")
        content.append(module_title)
        content.append("-" * len(module_title))
        content.append('\n'.join(doc.strip().splitlines()[2:]))
        content.append("")
        content.append(f".. automodule:: {module}")
        content.append("   :members:")
        content.append("   :member-order: bysource")
        content.append("")

    dynamic_fields_rst.write_text("\n".join(content), encoding="utf-8")


def suppress_function_headers():
    FunctionDocumenter.add_directive_header = lambda self, sig: None

def trim_module_docstring(app, what, name, obj, options, lines):
    if what != "module":
        return

    if not name.startswith("tests.test_routes_view.dynamic"):
        return

    if lines:
        lines[:] = []

def append_info_to_docstring(app, what, name, obj, options, lines):
    if what != "function":
        return

    # process only modules starting with tests.test_routes_view.dynamic
    if not getattr(obj, "__module__", "").startswith("tests.test_routes_view.dynamic"):
        return

    doc = inspect.getdoc(obj)
    lines.clear()

    if doc:
        doc_lines = doc.strip().splitlines()
        if doc_lines:
            title = doc_lines[0].strip()
            lines.append(title)
            lines.append("=" * len(title))
            lines.append("")

            rest_doc = doc_lines[1:]
            if rest_doc:
                lines += rest_doc
                lines.append("")

    for key, value in extract_all_variables_in_order(obj):
        lines.append("")
        if key == "url":
            lines.append("**Endpoint:**")
            lines.append("")
            lines.append(".. code-block:: bash")
            lines.append("")
            lines.append(f"   GET {value}")
        elif key in {"sql_query", "expected_sql_template"}:
            lines.append("**SQL query:**")
            lines.append("")
            lines.append(".. code-block:: sql")
            lines.append("")
            for line in value.splitlines():
                lines.append(f"   {line}")


def extract_all_variables_in_order(func):
    try:
        source = inspect.getsource(func)
        tree = ast.parse(source)
        results = []
        for node in tree.body:
            if isinstance(node, ast.FunctionDef):
                for sub in node.body:
                    if isinstance(sub, ast.Assign):
                        for target in sub.targets:
                            if isinstance(target, ast.Name) and target.id in {"url", "sql_query", "expected_sql_template"}:
                                value = sub.value
                                if isinstance(value, ast.Constant) and isinstance(value.value, str):
                                    results.append((target.id, value.value.strip()))
                                elif isinstance(value, ast.JoinedStr):
                                    parts = []
                                    for part in value.values:
                                        if isinstance(part, ast.Constant):
                                            parts.append(part.value)
                                        elif isinstance(part, ast.FormattedValue):
                                            parts.append("{...}")
                                    results.append((target.id, "".join(parts).strip()))
        return results
    except Exception as e:
        print(f"[sphinx][extract_all_variables_in_order] error: {e}")
        return []


# Copyright 2026 EcoFuture Technology Services LLC and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module for extracting filterable fields from route schemas.

This module defines the `RouteFilterFieldsService` class, which inspects FastAPI routes
based on `JsonapiRouteBase` to gather metadata about fields that can be used for filtering.

It supports:
- Native database fields (with type conversion to OpenAPI types).
- Related fields (resolved to list endpoints of related resources).
- Calculated fields (using annotated response types via `@calc_property`).

The main entry point is `RouteFilterFieldsService.get_fields(route_cls)`, which returns a list of `InputField` objects
representing filterable fields for the given route.
"""

from functools import cache
from typing import TYPE_CHECKING

from fastapi.routing import APIRoute

from pydantic import TypeAdapter

from bazis.core.routes_abstract.initial import InitialRouteBase
from bazis.core.schemas import CrudApiAction

from .schemas import InputField, RouteFilterFieldsSchemas


if TYPE_CHECKING:
    from .route_base import JsonapiRouteBase


class RouteFilterFieldsService:
    """
    Service that extracts filterable fields from route schema.

    Tags: RAG
    """

    @classmethod
    def get_fields(cls, route_cls: type['JsonapiRouteBase']) -> list[InputField]:
        """Return list of filterable fields for a given route class."""
        schema_factory = route_cls.schemas[CrudApiAction.LIST].schema_factory
        return [
            cls._serialize_field(route_schema_field, route_cls)
            for route_schema_field in schema_factory.fields_list
            if route_schema_field.can_filter is True
        ]

    @classmethod
    def _serialize_field(
        cls, route_schema_field: object, route_cls: type['JsonapiRouteBase']
    ) -> InputField:
        """
        Serialize a field to InputField format, determining the type (py_type)
        depending on whether it's a related field, a database field, or a calculated field.
        """
        field_info: dict[str, str] = {'name': route_schema_field.name}

        if route_schema_field.field_db_rel:
            # Related field — map to endpoint URL of related route.
            field_route_cls = cls._find_route_cls_for_model(
                route_schema_field.field_db_rel.related_model
            )
            field_info['py_type'] = (
                field_route_cls.url_path_for('action_list') if field_route_cls else 'unknown'
            )

        elif route_schema_field.field_db_attr:
            # Database attribute — resolve OpenAPI type via Python type.
            field_info['py_type'] = cls.get_openapi_type_name(route_schema_field.py_type)

        else:
            # Calculated field — get method's response_type and resolve its OpenAPI type.
            method = getattr(route_cls.model, route_schema_field.source, None)
            response_type = getattr(method, 'response_type', None)
            field_info['py_type'] = (
                cls.get_openapi_type_name(response_type) if response_type else 'unknown'
            )

        return InputField(**field_info)

    @staticmethod
    @cache
    def get_openapi_type_name(py_type: type) -> str:
        """
        Convert a Python type into a corresponding OpenAPI type name.
        If the type is a union (anyOf), returns the first non-null type.
        """
        schema = TypeAdapter(py_type).json_schema(schema_generator=RouteFilterFieldsSchemas)
        if 'type' in schema:
            return schema['type']
        if 'anyOf' in schema:
            return next(
                (
                    response_type['type']
                    for response_type in schema['anyOf']
                    if response_type['type'] != 'null'
                ),
                'unknown',
            )
        return 'unknown'

    @staticmethod
    @cache
    def _find_route_cls_for_model(model: type) -> type['JsonapiRouteBase'] | None:
        """
        Search registered FastAPI routes and return route class that corresponds to the given Django model.
        Only non-abstract InitialRouteBase subclasses are considered.
        """
        from bazis.core.app import app

        for route in app.router.routes:
            if (
                isinstance(route, APIRoute)
                and hasattr(route.endpoint, 'route_ctx')
                and issubclass(route.endpoint.route_ctx.route_cls, InitialRouteBase)
                and not getattr(route.endpoint.route_ctx.route_cls, 'abstract', False)
                and getattr(route.endpoint.route_ctx.route_cls, 'model', None) == model
            ):
                return route.endpoint.route_ctx.route_cls
        return None



# Copyright 2026 EcoFuture Technology Services LLC and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from django.apps import apps

from bazis.core.routes_abstract.jsonapi import JsonapiRouteBase
from bazis.core.schemas.fields import SchemaField, SchemaFields

from tests.utils.context_test_route_mixin import ContextTestRouteMixin


class ChildEntityRouteSet(JsonapiRouteBase):
    model = apps.get_model('entity.ChildEntity')

    fields = {
        None: SchemaFields(
            include={
                'parent_entities': None,
            },
        ),
    }


class DependentEntityRouteSet(JsonapiRouteBase):
    model = apps.get_model('entity.DependentEntity')


class DependentEntityNullRouteSet(JsonapiRouteBase):
    model = apps.get_model('entity.DependentEntityNull')


class ExtendedEntityRouteSet(JsonapiRouteBase):
    model = apps.get_model('entity.ExtendedEntity')


class ExtendedEntityNullRouteSet(JsonapiRouteBase):
    model = apps.get_model('entity.ExtendedEntityNull')


class WithProtectedEntityRouteSet(JsonapiRouteBase):
    model = apps.get_model('entity.WithProtectedEntity')


class ParentEntityRouteSet(JsonapiRouteBase):
    model = apps.get_model('entity.ParentEntity')

    # add fields (extended_entity, dependent_entities and calc property) to schema
    fields = {
        None: SchemaFields(
            include={
                'extended_entity': None,
                'dependent_entities': None,
                'active_children': SchemaField(source='active_children', required=False),
                'count_active_children': SchemaField(
                    source='count_active_children', required=False
                ),
                'has_inactive_children': SchemaField(
                    source='has_inactive_children', required=False
                ),
                'extended_entity_price': SchemaField(
                    source='extended_entity_price', required=False
                ),
            },
        ),
    }


class CountryRouteSet(JsonapiRouteBase):
    """Route for Country"""

    model = apps.get_model('entity.Country')


class VehicleModelRouteSet(JsonapiRouteBase):
    """Route for VehicleModel"""

    model = apps.get_model('entity.VehicleModel')


class VehicleBrandRouteSet(JsonapiRouteBase):
    """Route for VehicleBrand"""

    model = apps.get_model('entity.VehicleBrand')


class CarrierTaskRouteSet(JsonapiRouteBase):
    """Route for CarrierTask"""

    model = apps.get_model('entity.CarrierTask')


### DRIVER


class DriverRouteSet(JsonapiRouteBase):
    """Route for Driver"""

    model = apps.get_model('entity.Driver')


class DriverJsonHierarchyRouteSet(JsonapiRouteBase):
    """Route for Driver"""

    model = apps.get_model('entity.Driver')

    fields = {
        None: SchemaFields(
            include={
                'carrier_tasks_json_hierarchy': SchemaField(
                    source='carrier_tasks_json_hierarchy', required=False
                ),
            },
        ),
    }


class DriverContextRouteSet(ContextTestRouteMixin, JsonapiRouteBase):
    """Route for Driver"""

    model = apps.get_model('entity.Driver')

    fields = {
        None: SchemaFields(
            include={
                'carrier_tasks_context': SchemaField(
                    source='carrier_tasks_context', required=False
                ),
            },
        ),
    }


class DriverSliceRouteSet(JsonapiRouteBase):
    """Route for Driver"""

    model = apps.get_model('entity.Driver')

    fields = {
        None: SchemaFields(
            include={
                'last_trip': SchemaField(source='last_trip', required=False),
            },
        ),
    }


class DriverOrderByRouteSet(JsonapiRouteBase):
    """Route for Driver"""

    model = apps.get_model('entity.Driver')

    fields = {
        None: SchemaFields(
            include={
                'trips': SchemaField(source='trips', required=False),
            },
        ),
    }


### DIVISION


class DivisionJsonRouteSet(JsonapiRouteBase):
    """Route for Division"""

    model = apps.get_model('entity.Division')


class DivisonContextErrorRouteSet(ContextTestRouteMixin, JsonapiRouteBase):
    """Route for Driver"""

    model = apps.get_model('entity.Division')

    fields = {
        None: SchemaFields(
            include={
                # context that does not exist
                'drivers_list': SchemaField(source='drivers_list', required=False),
            },
        ),
    }


class DivisonFieldErrorRouteSet(JsonapiRouteBase):
    """Route for Driver"""

    model = apps.get_model('entity.Division')

    fields = {
        None: SchemaFields(
            include={
                # FieldJson - a field that does not exist
                'drivers_list1': SchemaField(source='drivers_list1', required=False),
            },
        ),
    }


### VEHICLE


class RelatedOneRelationMultipleFieldsRouteSet(JsonapiRouteBase):
    """Route for Related computed fields"""

    model = apps.get_model('entity.Vehicle')

    fields = {
        None: SchemaFields(
            include={
                'vehicle_model_info': SchemaField(source='vehicle_model_info', required=False),
            },
        ),
    }


class VehicleRelatedRouteSet(JsonapiRouteBase):
    """Route for Related computed fields"""

    model = apps.get_model('entity.Vehicle')

    fields = {
        None: SchemaFields(
            include={
                # FieldRelated - vehicle_capacity_1
                'vehicle_capacity_1': SchemaField(source='vehicle_capacity_1', required=False),
                # FieldRelated - vehicle_capacity_2
                'vehicle_capacity_2': SchemaField(source='vehicle_capacity_2', required=False),
                # FieldRelated - vehicle_capacity_3
                'vehicle_capacity_3': SchemaField(source='vehicle_capacity_3', required=False),
                # FieldRelated - brand, several fields
                'brand_info': SchemaField(source='brand_info', required=False),
                # FieldRelated - brand and country, several fields from several related tables
                'brand_and_country': SchemaField(source='brand_and_country', required=False),
            },
        ),
    }


class VehicleJsonRouteSet(JsonapiRouteBase):
    """Route for Json computed fields"""

    model = apps.get_model('entity.Vehicle')

    fields = {
        None: SchemaFields(
            include={
                # FieldJson - list of trips
                'carrier_task_list': SchemaField(source='carrier_task_list', required=False),
            },
        ),
    }


### IS EXISTS


class HasActiveTripRouteSet(JsonapiRouteBase):
    """Route for FieldIsExists computed fields"""

    model = apps.get_model('entity.Vehicle')

    fields = {
        None: SchemaFields(
            include={
                'has_active_trip': SchemaField(source='has_active_trip', required=False),
            },
        ),
    }


class VehicleExistsJoinRouteSet(JsonapiRouteBase):
    """Route for FieldIsExists computed fields"""

    model = apps.get_model('entity.Vehicle')

    fields = {
        None: SchemaFields(
            include={
                'has_active_trip_with_phone_join': SchemaField(
                    source='has_active_trip_with_phone_join', required=False
                ),
            },
        ),
    }


class VehicleExistsSubQueryRouteSet(JsonapiRouteBase):
    """Route for FieldIsExists computed fields"""

    model = apps.get_model('entity.Vehicle')

    fields = {
        None: SchemaFields(
            include={
                'has_active_trip_with_phone_subquery': SchemaField(
                    source='has_active_trip_with_phone_subquery', required=False
                ),
            },
        ),
    }


### AGGR SUBAGGR


class VehicleAggrRouteSet(JsonapiRouteBase):
    """Route for Aggr computed fields"""

    model = apps.get_model('entity.Vehicle')

    fields = {
        None: SchemaFields(
            include={
                'current_tasks_waste_weight_aggr': SchemaField(
                    source='current_tasks_waste_weight_aggr', required=False
                ),
            },
        ),
    }


class VehicleSubaggrRouteSet(JsonapiRouteBase):
    """Route for Subaggr computed fields"""

    model = apps.get_model('entity.Vehicle')

    fields = {
        None: SchemaFields(
            include={
                'current_tasks_waste_weight_subaggr': SchemaField(
                    source='current_tasks_waste_weight_subaggr', required=False
                ),
            },
        ),
    }



# Copyright 2026 EcoFuture Technology Services LLC and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from django.db.models import ForeignKey, ManyToManyField, Model, OneToOneField, QuerySet

from fastapi import Depends, Query

from bazis.core.routes_abstract.initial import inject_make
from bazis.core.schemas.enums import CrudApiAction
from bazis.core.services.route_ctx import get_route_ctx


class ServiceSparseFieldsets:
    """Service for handling JSON:API sparse fieldsets (fields[type] parameters)."""

    def __init__(
        self,
        route_ctx=Depends(get_route_ctx),
    ):
        self.route_ctx = route_ctx
        self.route_cls = route_ctx.route_cls
        self.fields = None

    @classmethod
    def get_fields_for_queryset(cls, queryset: QuerySet, inject) -> tuple:
        """Get field lists of different types for a model."""
        model_label = queryset.model.get_resource_label()

        k = f'fields_{model_label.replace(".", "_")}'
        v = getattr(inject, k, None)

        if v and isinstance(v, str):
            return cls.get_fields_for_inject_attribute_value(k, v, queryset, inject)
        return None, None, None

    @staticmethod
    def get_fields_for_inject_attribute_value(k: str, v: str, queryset: QuerySet, inject) -> tuple:
        """Parse field string into categorized field lists."""
        simple_fields_names = []
        calc_fields_names = []
        relation_fields_names = []

        v_splitted = [field.strip() for field in v.split(',')]
        setattr(inject, k, v_splitted)

        for field in v_splitted:
            if not hasattr(getattr(queryset.model, field), 'fields_calc'):
                if not isinstance(getattr(queryset.model, field), property):
                    simple_fields_names.append(field)
                    if hasattr(getattr(queryset.model, field), 'rel'):
                        relation_fields_names.append(field)
            else:
                calc_fields_names.append(field)
        return simple_fields_names, calc_fields_names, relation_fields_names

    @classmethod
    def _get_allowed_resource_types(cls, route_cls) -> tuple:
        """Get all allowed resource types by analyzing model relationships."""
        main_type = route_cls.model.get_resource_label()
        include_types = set()

        related_models = cls._get_related_models(route_cls.model)

        for related_model in related_models:
            if hasattr(related_model, 'get_resource_label'):
                related_type = related_model.get_resource_label()
                include_types.add(related_type)

        return main_type, sorted(include_types)

    @staticmethod
    def _get_related_models(model_class: type[Model]) -> set[type[Model]]:
        """Find all models related to the given model."""
        related_models = set()

        # Iterate through all model fields
        for field in model_class._meta.get_fields():
            # Direct relationships (ForeignKey, OneToOne)
            if isinstance(field, (ForeignKey, OneToOneField)):
                if field.related_model and field.related_model != model_class:
                    related_models.add(field.related_model)

            # Many-to-many relationships
            elif isinstance(field, ManyToManyField):
                if field.related_model and field.related_model != model_class:
                    related_models.add(field.related_model)

            # Reverse relationships (related_name)
            elif field.auto_created and not field.concrete:
                if field.related_model and field.related_model != model_class:
                    related_models.add(field.related_model)

        return related_models

    @staticmethod
    def create_inject_class(route_cls) -> tuple:
        """Factory method to create dynamic Inject class for sparse fieldsets."""
        main_type, include_types = ServiceSparseFieldsets._get_allowed_resource_types(route_cls)

        # 1. Class for LIST action (only main model)
        class_attrs_main = {
            '_inject_tags': {CrudApiAction.LIST},
            '__annotations__': {},
        }

        field_name_main = f'fields_{main_type}'.replace('.', '_')
        model_class_main = ServiceSparseFieldsets._get_model_for_type(
            main_type, route_cls, route_cls.model
        )

        if model_class_main:
            query_param_main = Query(
                default=None,
                alias=f'fields[{main_type}]',
            )
            class_attrs_main['__annotations__'][field_name_main] = str | None
            class_attrs_main[field_name_main] = query_param_main

        OnlyFieldsListClass = type('OnlyFieldsList', (), class_attrs_main)  # noqa: N806

        # 2. Class for RETRIEVE action (all types: main + included)
        class_attrs_retrieve = {
            '_inject_tags': {CrudApiAction.RETRIEVE},
            '__annotations__': {},
        }

        all_types = [main_type] + include_types

        for resource_type in all_types:
            field_name = f'fields_{resource_type}'.replace('.', '_')
            model_class = ServiceSparseFieldsets._get_model_for_type(
                resource_type, route_cls, route_cls.model
            )

            if model_class:
                query_param = Query(
                    default=None,
                    alias=f'fields[{resource_type}]',
                )
                class_attrs_retrieve['__annotations__'][field_name] = str | None
                class_attrs_retrieve[field_name] = query_param

        OnlyFieldsRetrieveClass = type('OnlyFieldsRetrieve', (), class_attrs_retrieve)  # noqa: N806

        # 3. Apply inject_make decorator
        return (
            inject_make(CrudApiAction.LIST)(OnlyFieldsListClass),
            inject_make(CrudApiAction.RETRIEVE)(OnlyFieldsRetrieveClass),
        )

    @staticmethod
    def _get_model_for_type(
        resource_type: str, route_cls, main_model: type[Model]
    ) -> type[Model] | None:
        """Find model class for a resource type by analyzing all related models."""

        # Check main model
        if resource_type == main_model.get_resource_label():
            return main_model

        # Search among related models
        related_models = ServiceSparseFieldsets._get_related_models(main_model)

        for related_model in related_models:
            if hasattr(related_model, 'get_resource_label'):
                if resource_type == related_model.get_resource_label():
                    return related_model

        return None

    @staticmethod
    def get_properties(model_cls: type[Model]) -> list[str]:
        """Get all property names from a model class and its ancestors."""
        props = {}
        for cls in model_cls.mro():
            for name, attr in vars(cls).items():
                if name != 'pk' and isinstance(attr, property):
                    props[name] = attr
        return list(props)



# Copyright 2026 EcoFuture Technology Services LLC and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any

from pydantic import (
    BaseModel,
    ValidationError,
    ValidationInfo,
)

from typing_extensions import deprecated

from bazis.core.models_abstract import InitialBase

from .utils import get_types


# check for nullable
def not_null_validator(cls: type[BaseModel], v: Any, field: ValidationInfo):
    """
    Validator to ensure that a field value is not null if the field is marked as
    non-nullable in the schema.

    Tags: RAG, INTERNAL
    """
    field_info = cls.model_fields[field.field_name]
    if field_info.json_schema_extra.get('nullable') is False and v is None:
        raise ValueError("Can't be null")
    return v


# check for blank
def not_blank_validator(cls: type[BaseModel], v: Any, field: ValidationInfo):
    """
    Validator to ensure that a field value is not blank if the field is marked as
    non-blank in the schema.

    Tags: RAG, INTERNAL
    """
    field_info = cls.model_fields[field.field_name]
    if field_info.json_schema_extra.get('blank') is False and not v:
        raise ValueError("Can't be blank")
    return v


@deprecated(
    'disabled because there are transition-type operations that require mandatory fields, which may well be read-only'
)
def readonly_validator(cls: type[BaseModel], values: Any):
    """
    Validator to remove read-only fields from the values if the schema action does
    not support read-only operations.
    """
    if not cls.schema_factory.api_action.for_read_only:
        for f_name, field_info in cls.model_fields.items():
            if field_info.json_schema_extra.get('readOnly'):
                values.pop(f_name, None)
    return values


def field_validate(cls: type[BaseModel], name, value):
    """
    Validate a field value against its schema, supporting nested types.

    Tags: RAG, INTERNAL
    """
    if field_info := cls.model_fields.get(name):
        types = get_types(field_info)
        if types:
            return model_validate(types[0], value, (name,))


def model_validate(model: type[BaseModel], value, loc: tuple):
    """
    Validate a model instance against its schema, raising a ValidationError with
    detailed context if validation fails.

    Tags: RAG, INTERNAL
    """
    try:
        return model.model_validate(value)
    except ValidationError as e:
        _id = None
        _type = None
        if isinstance(value, InitialBase):
            _id = value.pk
            _type = value.get_resource_label()
        elif isinstance(value, dict) and 'id' in value:
            _id = value['id']
            _type = value['type']

        for errs in e.errors():
            errs = errs if isinstance(errs, list) else [errs]
            for _e in errs:
                if isinstance(_e['loc'], tuple):
                    _e['loc'] = loc + _e['loc']
                else:
                    _e['loc'] = loc + (_e['loc'],)

                if 'ctx' not in _e:
                    _e['ctx'] = {}

                if _id:
                    _e['ctx']['_id'] = _id
                if _type:
                    _e['ctx']['_type'] = _type
        raise ValidationError.from_exception_data(str(e), e.errors()) from e



# Copyright 2026 EcoFuture Technology Services LLC and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Generated by Django 5.0.6 on 2024-06-27 14:35

from django.db import migrations, models

import bazis.core.fields


class Migration(migrations.Migration):

    dependencies = [
        ('entity', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='parententity',
            name='field',
            field=bazis.core.fields.ChoiceArrayField(base_field=models.CharField(choices=[('first_field', 'First field'), ('second_field', 'Second field'), ('third_field', 'Third field')], max_length=32), blank=True, null=True, size=None, verbose_name='Field'),
        ),
        migrations.AddField(
            model_name='parententity',
            name='state',
            field=models.CharField(choices=[('state_one', 'State one'), ('state_two', 'State two'), ('state_three', 'State three'), ('state_four', 'State four')], default='state_one', max_length=255, verbose_name='State'),
        ),
    ]



# Copyright 2026 EcoFuture Technology Services LLC and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
from bazis_test_utils.utils import get_api_client


@pytest.mark.django_db()
def test_filter_fields_calc_bool(sample_app):
    """
    Calculated field with Bool response.

    Calculated field for vehicles (Vehicle model) with information about whether the vehicle
    has an active trip (CarrierTask model).

    .. code-block:: python

        @calc_property(
            [
                FieldDynamic(
                    'carrier_tasks',
                    query=Q(dt_start__isnull=False) & ~Q(dt_finish__isnull=False),
                    alias='has_active_trip',
                ),
            ],
            as_filter=True,
        )
        def has_active_trip(self, dc: DependsCalc) -> bool:
            return dc.data.has_active_trip
    """

    url = '/api/v1/is_exists_one_relation/dynamic/vehicle/route_filter_fields/'
    response = get_api_client(sample_app).get(url)

    assert response.status_code == 200

    assert response.json() == {
        'fields': [
            {'name': 'dt_created', 'py_type': 'datetime'},
            {'name': 'dt_updated', 'py_type': 'datetime'},
            {'name': 'gnum', 'py_type': 'string'},
            {'name': 'vehicle_model', 'py_type': '/api/v1/dynamic/vehicle_model/'},
            {'name': 'has_active_trip', 'py_type': 'boolean'},
        ]
    }


@pytest.mark.django_db()
def test_filter_fields_calc_decimal(sample_app):
    """
    Calculated field with Decimal response.

    Calculated field for a vehicle (Vehicle model) with information about the load capacity of the corresponding model
    (VehicleModel model).

    .. code-block:: python

        @calc_property([FieldDynamic('vehicle_model')], as_filter=True)
        def vehicle_capacity(self, dc: DependsCalc) -> Decimal:
            return dc.data.vehicle_model.capacity
    """

    url = '/api/v1/related_one_relation_one_field/dynamic/vehicle/route_filter_fields/'
    response = get_api_client(sample_app).get(url)

    assert response.status_code == 200

    assert response.json() == {
        'fields': [
            {'name': 'dt_created', 'py_type': 'datetime'},
            {'name': 'dt_updated', 'py_type': 'datetime'},
            {'name': 'gnum', 'py_type': 'string'},
            {'name': 'vehicle_model', 'py_type': '/api/v1/dynamic/vehicle_model/'},
            {'name': 'vehicle_capacity', 'py_type': 'Decimal'},
        ]
    }


@pytest.mark.django_db()
def test_filter_fields_calc_dict(sample_app):
    """
    Calculated field with dict response.

    Calculated field for a vehicle (Vehicle model) with factory assembly data (VehicleAssembleInfo model).

    .. code-block:: python

        @calc_property([FieldDynamic('assemble_info')], as_filter=True)
        def vehicle_assemble_info(self, dc: DependsCalc) -> dict[str, Any]:
            return {
                'assembly_date': dc.data.assemble_info.assembly_date,
                'assembly_plant': dc.data.assemble_info.assembly_plant,
            }
    """

    url = '/api/v1/related_one_to_one/dynamic/vehicle/route_filter_fields/'
    response = get_api_client(sample_app).get(url)

    assert response.status_code == 200

    assert response.json() == {
        'fields': [
            {'name': 'dt_created', 'py_type': 'datetime'},
            {'name': 'dt_updated', 'py_type': 'datetime'},
            {'name': 'gnum', 'py_type': 'string'},
            {'name': 'vehicle_model', 'py_type': '/api/v1/dynamic/vehicle_model/'},
            {'name': 'vehicle_assemble_info', 'py_type': 'object'},
        ]
    }


@pytest.mark.django_db()
def test_filter_fields_calc_list(sample_app):
    """
    Calculated field with list response.

    Calculated field for vehicles (Vehicle model) with a list of trips (CarrierTask model).

    .. code-block:: python

        @calc_property(
            [
                FieldDynamic(
                    source='carrier_tasks',
                    fields=['id', 'dt_start', 'dt_finish', 'fact_waste_weight', 'driver_id'],
                ),
            ],
            as_filter=True,
        )
        def carrier_task_list(self, dc: DependsCalc) -> list:
            return [
                {
                    'id': task.id,
                    'dt_start': task.dt_start,
                    'dt_finish': task.dt_finish,
                    'fact_waste_weight': task.fact_waste_weight,
                    'driver_id': task.driver_id,
                }
                for task in dc.data.carrier_tasks
            ]
    """

    url = '/api/v1/json_one_relation/dynamic/vehicle/route_filter_fields/'

    response = get_api_client(sample_app).get(url)

    assert response.status_code == 200

    assert response.json() == {
        'fields': [
            {'name': 'dt_created', 'py_type': 'datetime'},
            {'name': 'dt_updated', 'py_type': 'datetime'},
            {'name': 'gnum', 'py_type': 'string'},
            {'name': 'vehicle_model', 'py_type': '/api/v1/dynamic/vehicle_model/'},
            {'name': 'carrier_task_list', 'py_type': 'array'},
        ]
    }


@pytest.mark.django_db()
def test_filter_fields_calc_anyof(sample_app):
    """
    Calculated field with anyOf response.

    Calculated field for a vehicle (Vehicle model) that sums the weight of all transported cargo (CarrierTask model).

    .. code-block:: python

        @calc_property(
            [
                FieldDynamic(
                    source='carrier_tasks__fact_waste_weight',
                    func='Sum',
                    alias='finished_tasks_waste_weight',
                    query=Q(dt_finish__isnull=False),
                ),
            ]
            , as_filter=True
        )
        def finished_tasks_waste_weight(self, dc: DependsCalc) -> Decimal | None:
            return dc.data.finished_tasks_waste_weight
    """

    url = '/api/v1/subaggr/dynamic/vehicle/route_filter_fields/'

    response = get_api_client(sample_app).get(url)

    assert response.status_code == 200

    assert response.json() == {
        'fields': [
            {'name': 'dt_created', 'py_type': 'datetime'},
            {'name': 'dt_updated', 'py_type': 'datetime'},
            {'name': 'gnum', 'py_type': 'string'},
            {'name': 'vehicle_model', 'py_type': '/api/v1/dynamic/vehicle_model/'},
            {'name': 'finished_tasks_waste_weight', 'py_type': 'Decimal'},
        ]
    }
